import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit

# from species.data.database import Database
from species.read.read_color import ReadColorColor, ReadColorMagnitude
from species.util import test_util


class TestColor:
    def setup_class(self):
        self.limit = 1e-8

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    # def test_read_color_magnitude(self):
    #     database = Database()
    #     database.add_photometry("vlm-plx")
    #     database.add_photometry("leggett")
    #
    #     read_colormag = ReadColorMagnitude(
    #         "vlm-plx", ("MKO/NSFCam.J", "MKO/NSFCam.H"), "MKO/NSFCam.J"
    #     )
    #
    #     assert read_colormag.filters_color == ("MKO/NSFCam.J", "MKO/NSFCam.H")
    #     assert read_colormag.filter_mag == "MKO/NSFCam.J"
    #
    # def test_get_color_magnitude(self):
    #     read_colormag = ReadColorMagnitude(
    #         "leggett", ("MKO/NSFCam.J", "MKO/NSFCam.H"), "MKO/NSFCam.J"
    #     )
    #
    #     colormag_box = read_colormag.get_color_magnitude(object_type=None)
    #     colormag_box.open_box()
    #
    #     assert np.nansum(colormag_box.color) == pytest.approx(14.05)
    #     assert np.nansum(colormag_box.magnitude) == pytest.approx(1519.373)
    #
    # def test_read_color_color(self):
    #     read_colorcolor = ReadColorColor(
    #         "vlm-plx",
    #         (("MKO/NSFCam.J", "MKO/NSFCam.H"), ("MKO/NSFCam.H", "MKO/NSFCam.K")),
    #     )
    #
    #     assert read_colorcolor.filters_colors == (
    #         ("MKO/NSFCam.J", "MKO/NSFCam.H"),
    #         ("MKO/NSFCam.H", "MKO/NSFCam.K"),
    #     )
    #
    # def test_get_color_color(self):
    #     read_colorcolor = ReadColorColor(
    #         "vlm-plx",
    #         (("MKO/NSFCam.J", "MKO/NSFCam.H"), ("MKO/NSFCam.H", "MKO/NSFCam.K")),
    #     )
    #
    #     colorcolor_box = read_colorcolor.get_color_color(object_type=None)
    #
    #     assert np.nansum(colorcolor_box.color1) == pytest.approx(166.99384)
    #     assert np.nansum(colorcolor_box.color2) == pytest.approx(138.5127)



"""
Module with utility functions for plotting data.
"""

import warnings

from string import ascii_lowercase
from typing import Dict, List, Optional, Tuple

import numpy as np

from typeguard import typechecked

from species.core import constants
from species.util.model_util import convert_model_name


@typechecked
def sptype_to_index(
    field_range: Tuple[str, str], spec_types: np.ndarray, check_subclass: bool
) -> np.ndarray:
    """
    Function for mapping the spectral types of stellar and
    substellar objects to indices that corresponds with the
    discrete colorbar of a color-magnitude or color-color
    diagram.

    Parameters
    ----------
    field_range : tuple(str, str)
        Range of the discrete colorbar for the field objects. The tuple
        should contain the lower and upper value ('early M', 'late M',
        'early L', 'late L', 'early T', 'late T', 'early Y). Also
        stellar spectral types can be specified.
    spec_types : np.ndarray
        Array with the spectral types.
    check_subclass : bool
        Set to ``True`` if the discrete colorbar should distinguish
        early and late spectral types with different colors or set
        to ``False`` if subclasses should not be distinguished.

    Returns
    -------
    np.ndarray
        Array with spectral types mapped to indices. Spectral types
        that are not within the range specified with ``field_range``
        will be set to NaN.
    """

    spt_discrete = np.zeros(spec_types.size)

    if check_subclass:
        spt_check = [
            ("early O", "late O"),
            ("early B", "late B"),
            ("early A", "late A"),
            ("early F", "late F"),
            ("early G", "late G"),
            ("early K", "late K"),
            ("early M", "late M"),
            ("early L", "late L"),
            ("early T", "late T"),
            ("early Y", "late Y"),
        ]

        for i, item in enumerate(spec_types):
            if isinstance(item, bytes):
                # Catch SpT names that were read from FITS files
                # and stored as bytes in the HDF5 database
                item = item.decode("utf-8")

            if item[0:2] in ["O0", "O1", "O2", "O3", "O4"]:
                spt_discrete[i] = 0.5

            elif item[0:2] in ["O5", "O6", "O7", "O8", "O9"]:
                spt_discrete[i] = 1.5

            elif item[0:2] in ["B0", "B1", "B2", "B3", "B4"]:
                spt_discrete[i] = 2.5

            elif item[0:2] in ["B5", "B6", "B7", "B8", "B9"]:
                spt_discrete[i] = 3.5

            elif item[0:2] in ["A0", "A1", "A2", "A3", "A4"]:
                spt_discrete[i] = 4.5

            elif item[0:2] in ["A5", "A6", "A7", "A8", "A9"]:
                spt_discrete[i] = 5.5

            elif item[0:2] in ["F0", "F1", "F2", "F3", "F4"]:
                spt_discrete[i] = 6.5

            elif item[0:2] in ["F5", "F6", "F7", "F8", "F9"]:
                spt_discrete[i] = 7.5

            elif item[0:2] in ["G0", "G1", "G2", "G3", "G4"]:
                spt_discrete[i] = 8.5

            elif item[0:2] in ["G5", "G6", "G7", "G8", "G9"]:
                spt_discrete[i] = 9.5

            elif item[0:2] in ["K0", "K1", "K2", "K3", "K4"]:
                spt_discrete[i] = 10.5

            elif item[0:2] in ["K5", "K6", "K7", "K8", "K9"]:
                spt_discrete[i] = 11.5

            elif item[0:2] in ["M0", "M1", "M2", "M3", "M4"]:
                spt_discrete[i] = 12.5

            elif item[0:2] in ["M5", "M6", "M7", "M8", "M9"]:
                spt_discrete[i] = 13.5

            elif item[0:2] in ["L0", "L1", "L2", "L3", "L4"]:
                spt_discrete[i] = 14.5

            elif item[0:2] in ["L5", "L6", "L7", "L8", "L9"]:
                spt_discrete[i] = 15.5

            elif item[0:2] in ["T0", "T1", "T2", "T3", "T4"]:
                spt_discrete[i] = 16.5

            elif item[0:2] in ["T5", "T6", "T7", "T8", "T9"]:
                spt_discrete[i] = 17.5

            elif "Y" in item:
                spt_discrete[i] = 18.5

            else:
                spt_discrete[i] = np.nan

        count = 0
        for i, item in enumerate(spt_check):
            for j in range(2):
                if field_range[0] == item[j]:
                    spt_discrete -= float(count)
                    break
                count += 1

    else:
        spt_check = ["O", "B", "A", "F", "G", "K", "M", "L", "T", "Y"]

        for i, item in enumerate(spec_types):
            if isinstance(item, bytes):
                # Catch SpT names that were read from FITS files
                # and stored as bytes in the HDF5 database
                item = item.decode("utf-8")

            if item[0] == "O":
                spt_discrete[i] = 0.5

            elif item[0] == "B":
                spt_discrete[i] = 1.5

            elif item[0] == "A":
                spt_discrete[i] = 2.5

            elif item[0] == "F":
                spt_discrete[i] = 3.5

            elif item[0] == "G":
                spt_discrete[i] = 4.5

            elif item[0] == "K":
                spt_discrete[i] = 5.5

            elif item[0] == "M":
                spt_discrete[i] = 6.5

            elif item[0] == "L":
                spt_discrete[i] = 7.5

            elif item[0] == "T":
                spt_discrete[i] = 8.5

            elif item[0] == "Y":
                spt_discrete[i] = 9.5

            else:
                spt_discrete[i] = np.nan

        for i, item in enumerate(spt_check):
            if field_range[0] == item:
                spt_discrete -= float(i)
                break

    set_to_nan = spt_discrete < 0.0
    spt_discrete[set_to_nan] = np.nan

    return spt_discrete


@typechecked
def update_labels(param: List[str], object_type: str = "planet") -> List[str]:
    """
    Function for formatting the model parameters to use them
    as labels in the posterior plot.

    Parameters
    ----------
    param : list
        List with names of the model parameters.
    object_type : str
        Object type ('planet' or 'star'). With 'planet', the radius
        and mass are expressed in Jupiter units. With 'star', the
        radius and mass are expressed in solar units.

    Returns
    -------
    list
        List with parameter labels for plots.
    """

    cloud_species = ["Fe", "MgSiO3", "Al2O3", "Na2S", "KCl"]

    cloud_labels = ["Fe", r"MgSiO_{3}", r"Al_{2}O_{3}", r"Na_{2}S", "KCl"]

    abund_species = [
        "CO_all_iso",
        "CO_all_iso_HITEMP",
        "H2O",
        "H2O_HITEMP",
        "H2O_main_iso",
        "CH4",
        "CH4_main_iso",
        "NH3",
        "NH3_main_iso",
        "CO2",
        "CO2_main_iso",
        "H2S",
        "H2S_main_iso",
        "Na",
        "Na_allard",
        "Na_burrows",
        "Na_lor_cur",
        "K",
        "K_allard",
        "K_burrows",
        "K_lor_cur",
        "PH3",
        "PH3_main_iso",
        "VO",
        "VO_Plez",
        "TiO",
        "TiO_all_Exomol",
        "TiO_all_iso_Plez",
        "FeH",
        "FeH_main_iso",
        "MgSiO3(c)",
        "Fe(c)",
        "Al2O3(c)",
        "Na2S(c)",
        "KCL(c)",
    ]

    abund_labels = [
        "CO",
        "CO",
        "H_{2}O",
        "H_{2}O",
        "H_{2}O",
        "CH_{4}",
        "CH_{4}",
        "NH_{3}",
        "NH_{3}",
        "CO_{2}",
        "CO_{2}",
        "H_{2}S",
        "H_{2}S",
        "Na",
        "Na",
        "Na",
        "Na",
        "K",
        "K",
        "K",
        "K",
        "PH_{3}",
        "PH_{3}",
        "VO",
        "VO",
        "TiO",
        "TiO",
        "TiO",
        "FeH",
        "FeH",
        "MgSiO_{3}",
        "Fe",
        "Al_{2}O_{3}",
        "Na_{2}S",
        "KCl",
    ]

    if "teff" in param:
        index = param.index("teff")
        param[index] = r"$T_\mathrm{eff}$ (K)"

    for i in range(100):
        if f"disk_teff_{i}" in param:
            index = param.index(f"disk_teff_{i}")
            param[index] = r"$T_\mathrm{disk,}$" + rf"$_\mathrm{{{i+1}}}$ (K)"

        else:
            break

    if "logg" in param:
        index = param.index("logg")
        param[index] = r"$\log\,g$"

    if "logg_0" in param:
        index = param.index("logg_0")
        param[index] = r"$\log\,g_\mathrm{1}$"

    if "logg_1" in param:
        index = param.index("logg_1")
        param[index] = r"$\log\,g_\mathrm{2}$"

    if "metallicity" in param:
        index = param.index("metallicity")
        param[index] = "[Fe/H]"

    if "feh" in param:
        index = param.index("feh")
        param[index] = "[Fe/H]"

    if "feh_0" in param:
        index = param.index("feh_0")
        param[index] = r"[Fe/H]$_\mathrm{1}$"

    if "feh_1" in param:
        index = param.index("feh_1")
        param[index] = r"[Fe/H]$_\mathrm{2}$"

    if "fsed" in param:
        index = param.index("fsed")
        param[index] = r"$f_\mathrm{sed}$"

    if "fsed_0" in param:
        index = param.index("fsed_0")
        param[index] = r"$f_\mathrm{sed,1}$"

    if "fsed_1" in param:
        index = param.index("fsed_1")
        param[index] = r"$f_\mathrm{sed,2}$"

    if "vsini" in param:
        index = param.index("vsini")
        param[index] = r"$v\,\sin\,i$ (km s$^{-1}$)"

    if "vsini_0" in param:
        index = param.index("vsini_0")
        param[index] = r"$v\,\sin\,i_\mathrm{1}$ (km s$^{-1}$)"

    if "vsini_1" in param:
        index = param.index("vsini_1")
        param[index] = r"$v\,\sin\,i_\mathrm{2}$ (km s$^{-1}$)"

    if "rad_vel" in param:
        index = param.index("rad_vel")
        param[index] = r"$v_\mathrm{r}$ (km s$^{-1}$)"

    if "rad_vel_0" in param:
        index = param.index("rad_vel_0")
        param[index] = r"$v_\mathrm{r,1}$ (km s$^{-1}$)"

    if "rad_vel_1" in param:
        index = param.index("rad_vel_1")
        param[index] = r"$v_\mathrm{r,2}$ (km s$^{-1}$)"

    if "log_lum_0" in param:
        index = param.index("log_lum_0")
        param[index] = r"$\log\,L_\mathrm{1}/L_\mathrm{\odot}$"

    if "log_lum_1" in param:
        index = param.index("log_lum_1")
        param[index] = r"$\log\,L_\mathrm{2}/L_\mathrm{\odot}$"

    if "f_clouds" in param:
        index = param.index("f_clouds")
        param[index] = r"$w_\mathrm{clouds}$"

    if "c_o_ratio" in param:
        index = param.index("c_o_ratio")
        param[index] = r"C/O"

    if "radius" in param:
        index = param.index("radius")
        if object_type == "planet":
            param[index] = r"$R_\mathrm{p}$ ($R_\mathrm{J}$)"
        elif object_type == "star":
            param[index] = r"$R_\ast$ ($R_\mathrm{\odot}$)"

    if "distance" in param:
        index = param.index("distance")
        param[index] = "$d$ (pc)"

    if "parallax" in param:
        index = param.index("parallax")
        param[index] = r"$\varpi$ (mas)"

    if "parallax_0" in param:
        index = param.index("parallax_0")
        param[index] = r"$\varpi_\mathrm{1}$ (mas)"

    if "parallax_1" in param:
        index = param.index("parallax_1")
        param[index] = r"$\varpi_\mathrm{2}$ (mas)"

    if "mass" in param:
        index = param.index("mass")
        if object_type == "planet":
            param[index] = r"$M$ ($M_\mathrm{J}$)"
        elif object_type == "star":
            param[index] = r"$M_\ast$ ($M_\mathrm{\odot}$)"

    if "mass_0" in param:
        index = param.index("mass_0")
        if object_type == "planet":
            param[index] = r"$M_\mathrm{1}$ ($M_\mathrm{J}$)"
        elif object_type == "star":
            param[index] = r"$M_\mathrm{1}$ ($M_\mathrm{\odot}$)"

    if "mass_1" in param:
        index = param.index("mass_1")
        if object_type == "planet":
            param[index] = r"$M_\mathrm{2}$ ($M_\mathrm{J}$)"
        elif object_type == "star":
            param[index] = r"$M_\mathrm{2}$ ($M_\mathrm{\odot}$)"

    # for i, item in enumerate(ascii_lowercase[1:]):
    #     if f"mass_{i}" in param:
    #         index = param.index(f"mass_{i}")
    #         param[index] = rf"$M_\mathrm{{{item}}}$ ($M_\mathrm{{J}}$)"
    #     else:
    #         break

    if "log_mass" in param:
        index = param.index("log_mass")
        if object_type == "planet":
            param[index] = r"$\log\,M/M_\mathrm{J}$"
        elif object_type == "star":
            param[index] = r"$\log\,M_\ast/M_\mathrm{\odot}$"

    if "log_mass_0" in param:
        index = param.index("log_mass_0")
        if object_type == "planet":
            param[index] = r"$\log\,M_\mathrm{1}/M_\mathrm{J}$"
        elif object_type == "star":
            param[index] = r"$\log\,M_\mathrm{1}/M_\mathrm{\odot}$"

    if "log_mass_1" in param:
        index = param.index("log_mass_1")
        if object_type == "planet":
            param[index] = r"$\log\,M_\mathrm{2}/M_\mathrm{J}$"
        elif object_type == "star":
            param[index] = r"$\log\,M_\mathrm{2}/M_\mathrm{\odot}$"

    if "mass_ratio" in param:
        index = param.index("mass_ratio")
        param[index] = r"$q$"

    if "age" in param:
        index = param.index("age")
        param[index] = "Age (Myr)"

    # if "mass_1" in param:
    #     index = param.index("mass_1")
    #     param[index] = r"$M_\mathrm{b}$ ($M_\mathrm{J}$)"
    #
    # if "mass_2" in param:
    #     index = param.index("mass_2")
    #     param[index] = r"$M_\mathrm{c}$ ($M_\mathrm{J}$)"

    if "entropy" in param:
        index = param.index("entropy")
        param[index] = r"$S_\mathrm{i}$ ($k_\mathrm{B}/\mathrm{baryon}$)"

    if "entropy_1" in param:
        index = param.index("entropy_1")
        param[index] = r"$S_\mathrm{i,b}$ ($k_\mathrm{B}/\mathrm{baryon}$)"

    if "entropy_2" in param:
        index = param.index("entropy_2")
        param[index] = r"$S_\mathrm{i,c}$ ($k_\mathrm{B}/\mathrm{baryon}$)"

    if "dfrac_1" in param:
        index = param.index("dfrac_1")
        param[index] = r"$\log\,D_\mathrm{i,b}$"

    if "dfrac_2" in param:
        index = param.index("dfrac_2")
        param[index] = r"$\log\,D_\mathrm{i,c}$"

    if "y_frac" in param:
        index = param.index("y_frac")
        param[index] = r"$Y$"

    if "yfrac_1" in param:
        index = param.index("yfrac_1")
        param[index] = r"$Y_\mathrm{b}$"

    if "yfrac_2" in param:
        index = param.index("yfrac_2")
        param[index] = r"$Y_\mathrm{c}$"

    if "mcore_1" in param:
        index = param.index("mcore_1")
        param[index] = r"$M_\mathrm{core,b}$ ($M_\mathrm{E}$)"

    if "mcore_2" in param:
        index = param.index("mcore_2")
        param[index] = r"$M_\mathrm{core,c}$ ($M_\mathrm{E}$)"

    for i, item in enumerate(ascii_lowercase[1:]):
        if f"teff_evol_{i}" in param:
            index = param.index(f"teff_evol_{i}")
            param[index] = rf"$T_\mathrm{{eff,{item}}}$ (K)"
        else:
            break

    for i, item in enumerate(ascii_lowercase[1:]):
        if f"radius_evol_{i}" in param:
            index = param.index(f"radius_evol_{i}")
            param[index] = rf"$R_\mathrm{{{item}}}$ ($R_\mathrm{{J}}$)"
        else:
            break

    for i, item in enumerate(ascii_lowercase[1:]):
        if f"logg_evol_{i}" in param:
            index = param.index(f"logg_evol_{i}")
            param[index] = rf"$\log\,g_\mathrm{{{item}}}$"
        else:
            break

    for i, item in enumerate(ascii_lowercase[1:]):
        if f"inflate_lbol{i}" in param:
            index = param.index(f"inflate_lbol{i}")
            param[index] = rf"$\sigma_{{L,{{{item}}}}}$ (dex)"
        else:
            break

    for i, item in enumerate(ascii_lowercase[1:]):
        if f"inflate_mass{i}" in param:
            index = param.index(f"inflate_mass{i}")
            param[index] = rf"$\sigma_{{M,{{{item}}}}}$ ($M_\mathrm{{J}}$)"
        else:
            break

    if "luminosity" in param:
        index = param.index("luminosity")
        if object_type == "planet":
            param[index] = r"$\log\,L/L_\mathrm{\odot}$"
        elif object_type == "star":
            param[index] = r"$\log\,L_\ast/L_\mathrm{\odot}$"

    if "luminosity_ratio" in param:
        index = param.index("luminosity_ratio")
        param[index] = r"$\log\,L_\mathrm{1}/L_\mathrm{2}$"

    if "luminosity_disk_planet" in param:
        index = param.index("luminosity_disk_planet")
        if object_type == "planet":
            param[index] = r"$L_\mathrm{disk}/L_\mathrm{atm}$"
        elif object_type == "star":
            param[index] = r"$L_\mathrm{disk}/L_\ast$"

    if "log_lum" in param:
        index = param.index("log_lum")
        if object_type == "planet":
            param[index] = r"$\log\,L/L_\mathrm{\odot}$"
        elif object_type == "star":
            param[index] = r"$\log\,L_\ast/L_\mathrm{\odot}$"

    if "log_lum_atm" in param:
        index = param.index("log_lum_atm")
        if object_type == "planet":
            param[index] = r"$\log\,L_\mathrm{p}/L_\mathrm{\odot}$"
        elif object_type == "star":
            param[index] = r"$\log\,L_\ast/L_\mathrm{\odot}$"

    if "log_lum_disk" in param:
        index = param.index("log_lum_disk")
        param[index] = r"$\log\,L_\mathrm{disk}/L_\mathrm{\odot}$"

    if "lognorm_radius" in param:
        index = param.index("lognorm_radius")
        param[index] = r"$\log\,r_\mathrm{g}$"

    if "lognorm_sigma" in param:
        index = param.index("lognorm_sigma")
        param[index] = r"$\sigma_\mathrm{g}$"

    if "lognorm_ext" in param:
        index = param.index("lognorm_ext")
        param[index] = r"$A_V$"

    if "powerlaw_min" in param:
        index = param.index("powerlaw_min")
        param[index] = r"$\log\,a_\mathrm{min}/\mathrm{µm}$"

    if "powerlaw_max" in param:
        index = param.index("powerlaw_max")
        param[index] = r"$\log\,a_\mathrm{max}/\mathrm{µm}$"

    if "powerlaw_exp" in param:
        index = param.index("powerlaw_exp")
        param[index] = r"$\beta$"

    if "powerlaw_ext" in param:
        index = param.index("powerlaw_ext")
        param[index] = r"$A_V$"

    if "ism_ext" in param:
        index = param.index("ism_ext")
        param[index] = r"$A_V$"

    if "ism_ext_0" in param:
        index = param.index("ism_ext_0")
        param[index] = r"$A_{V,\mathrm{1}}$"

    if "ism_ext_1" in param:
        index = param.index("ism_ext_1")
        param[index] = r"$A_{V,\mathrm{2}}$"

    if "ism_red" in param:
        index = param.index("ism_red")
        param[index] = r"$R_V$"

    if "ext_av" in param:
        index = param.index("ext_av")
        param[index] = r"$A_V$"

    if "ext_av_0" in param:
        index = param.index("ext_av_0")
        param[index] = r"$A_{V,\mathrm{1}}$"

    if "ext_av_1" in param:
        index = param.index("ext_av_1")
        param[index] = r"$A_{V,\mathrm{2}}$"

    if "ext_rv" in param:
        index = param.index("ext_rv")
        param[index] = r"$R_V$"

    for item in param:
        if item.startswith("phot_ext_"):
            index = param.index(item)
            filter_name = item[9:].split("/")[1]
            param[index] = rf"$A_\mathrm{{{filter_name}}}$"
            break

    if "tint" in param:
        index = param.index("tint")
        param[index] = r"$T_\mathrm{int}$ (K)"

    for i in range(15):
        if f"t{i}" in param:
            index = param.index(f"t{i}")
            param[index] = rf"$T_\mathrm{{{i}}}$ (K)"

    if "alpha" in param:
        index = param.index("alpha")
        param[index] = r"$\alpha$"

    if "log_sigma_alpha" in param:
        index = param.index("log_sigma_alpha")
        param[index] = r"$\log\,\sigma_\alpha$"

    if "log_delta" in param:
        index = param.index("log_delta")
        param[index] = r"$\log\,\delta$"

    if "T_bottom" in param:
        index = param.index("T_bottom")
        param[index] = r"$T_\mathrm{bot}$ (K)"

    num_layer = 6  # could make a variable in the future
    for i in range(num_layer):
        if f"PTslope_{num_layer - i}" in param:
            index = param.index(f"PTslope_{num_layer - i}")
            param[index] = rf"dlnT/dlnP$_{i}$ "

    if "log_p_quench" in param:
        index = param.index("log_p_quench")
        param[index] = r"$\log\,P_\mathrm{quench}$"

    if "sigma_lnorm" in param:
        index = param.index("sigma_lnorm")
        param[index] = r"$\sigma_\mathrm{g}$"

    if "log_kzz" in param:
        index = param.index("log_kzz")
        param[index] = r"$\log\,K_\mathrm{zz}$"

    if "kzz" in param:
        # Backward compatibility
        index = param.index("kzz")
        param[index] = r"$\log\,K_\mathrm{zz}$"

    if "log_co_iso" in param:
        index = param.index("log_co_iso")
        param[index] = r"$\log\,^{12}\mathrm{CO}/^{13}\mathrm{CO}$"

        print(cloud_species)

    for item in param:

        if item[-9:] == "_fraction":
            from petitRADTRANS.chemistry.utils import simplify_species_list

            index = param.index(item)
            cloud_simple = simplify_species_list([item[:-9]])[0]
            param[index] = rf"$\log\,\tilde{{\mathrm{{X}}}}_\mathrm{{{cloud_simple}}}$"

        if item[-4:] == "_tau":
            from petitRADTRANS.chemistry.utils import simplify_species_list

            index = param.index(item)
            cloud_simple = simplify_species_list([item[:-4]])[0]
            param[index] = rf"$\bar{{\tau}}_\mathrm{{{cloud_simple}}}$"

        if item[:11] == "log_p_base_":
            from petitRADTRANS.chemistry.utils import simplify_species_list

            index = param.index(item)
            cloud_simple = simplify_species_list([item[11:-3]])[0]
            param[index] = rf"$\log\,P_\mathrm{{{cloud_simple}}}$"

        if item[:5] == "fsed_":
            from petitRADTRANS.chemistry.utils import simplify_species_list

            index = param.index(item)
            cloud_simple = simplify_species_list([item[5:-3]])[0]
            param[index] = rf"fsed$_\mathrm{{{cloud_simple}}}$"

    for i, item_i in enumerate(cloud_species):
        for j, item_j in enumerate(cloud_species):
            if f"{item_i.lower()}_{item_j.lower()}_ratio" in param:
                index = param.index(f"{item_i.lower()}_{item_j.lower()}_ratio")
                param[index] = (
                    rf"$\log\,\tilde{{\mathrm{{X}}}}"
                    rf"_\mathrm{{{cloud_labels[i]}}}/"
                    rf"\mathrm{{\tilde{{X}}}}_\mathrm{{{cloud_labels[j]}}}$"
                )

    for i, item in enumerate(abund_species):
        if item in param:
            index = param.index(item)
            param[index] = rf"$\log\,\mathrm{{{abund_labels[i]}}}$"

    for i, item in enumerate(param):
        if item[0:8] == "scaling_":
            item_name = item[8:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            if item_name.find("/") > 0:
                item_name = item_name[item_name.find("/") + 1 :]
            param[i] = rf"$a_\mathrm{{{item_name}}}$"

        elif item[0:6] == "error_":
            item_name = item[6:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            if item_name.find("/") > 0:
                item_name = item_name[item_name.find("/") + 1 :]
            param[i] = rf"$b_\mathrm{{{item_name}}}$"

        elif item[0:10] == "log_error_":
            item_name = item[10:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            if item_name.find("/") > 0:
                item_name = item_name[item_name.find("/") + 1 :]
            param[i] = rf"$\log\,b_\mathrm{{{item_name}}}$"

        elif item[0:7] == "radvel_":
            item_name = item[7:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            param[i] = rf"$v_\mathrm{{r,{item_name}}}$ (km s$^{{-1}}$)"

        elif item[0:8] == "rad_vel_":
            item_name = item[8:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            param[i] = rf"$v_\mathrm{{r,{item_name}}}$ (km s$^{{-1}}$)"

        elif item[0:6] == "vsini_":
            item_name = item[6:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            param[i] = rf"$v\,\sin\,i_\mathrm{{{item_name}}}$ (km s$^{-1}$)"

        elif item[0:11] == "wavelength_":
            item_name = item[11:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            param[i] = rf"$c_\mathrm{{{item_name}}}$ (nm)"

        elif item[:6] == "error_":
            item_name = item[6:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            if item_name.find("/") > 0:
                item_name = item_name[item_name.find("/") + 1 :]
            param[i] = rf"$b_\mathrm{{{item_name}}}$"

        elif item[:10] == "log_error_":
            item_name = item[10:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            if item_name.find("/") > 0:
                item_name = item_name[item_name.find("/") + 1 :]
            param[i] = rf"$\log\,b_\mathrm{{{item_name}}}$"

        elif item[0:9] == "corr_len_":
            item_name = item[9:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            param[i] = rf"$\log\,\ell_\mathrm{{{item_name}}}$"

        elif item[0:9] == "corr_amp_":
            item_name = item[9:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
            param[i] = rf"$f_\mathrm{{{item_name}}}$"

        elif item[0:6] == "ratio_":
            item_name = item[6:]
            if item_name.find("\\_") == -1 and item_name.find("_") > 0:
                item_name = item_name.replace("_", "\\_")
                item_name = item_name.split("/")[1]
            param[i] = rf"$r_\mathrm{{{item_name}}}$"

    if "c_h_ratio" in param:
        index = param.index("c_h_ratio")
        param[index] = r"[C/H]"

    if "o_h_ratio" in param:
        index = param.index("o_h_ratio")
        param[index] = r"[O/H]"

    for i in range(100):
        if f"teff_{i}" in param:
            index = param.index(f"teff_{i}")
            param[index] = rf"$T_\mathrm{{eff,{i+1}}}$ (K)"

        else:
            break

    for i in range(100):
        if f"radius_{i}" in param:
            index = param.index(f"radius_{i}")
            if object_type == "planet":
                param[index] = rf"$R_\mathrm{{{i+1}}}$ ($R_\mathrm{{J}}$)"
            elif object_type == "star":
                param[index] = rf"$R_\mathrm{{{i+1}}}$ ($R_\mathrm{{\odot}}$)"

        else:
            break

    for i in range(100):
        if f"luminosity_{i}" in param:
            index = param.index(f"luminosity_{i}")
            param[index] = rf"$\log\,L_\mathregular{{{i+1}}}/L_\mathregular{{\odot}}$"

        else:
            break

    if "disk_teff" in param:
        index = param.index("disk_teff")
        param[index] = r"$T_\mathrm{disk}$ (K)"

    for i in range(100):
        if f"disk_teff_{i}" in param:
            index = param.index(f"disk_teff_{i}")
            param[index] = r"$T_\mathrm{disk,}$" + rf"$_\mathrm{{{i+1}}}$ (K)"

        else:
            break

    if "disk_radius" in param:
        index = param.index("disk_radius")
        if object_type == "planet":
            param[index] = r"$R_\mathrm{disk}$ ($R_\mathrm{J}$)"
        elif object_type == "star":
            param[index] = r"$R_\mathrm{disk}$ (au)"

    for i in range(100):
        if f"disk_radius_{i}" in param:
            index = param.index(f"disk_radius_{i}")
            if object_type == "planet":
                param[index] = (
                    r"$R_\mathrm{disk,}$" + rf"$_{{{i+1}}}$ ($R_\mathrm{{J}}$)"
                )
            elif object_type == "star":
                param[index] = r"$R_\mathrm{disk,}$" + rf"$_\mathrm{{{i+1}}}$ (au)"

        else:
            break

    if "radius_bb" in param:
        index = param.index("radius_bb")
        if object_type == "planet":
            param[index] = r"$R_\mathrm{bb}$ ($R_\mathrm{J}$)"
        elif object_type == "star":
            param[index] = r"$R_\mathrm{bb}$ (au)"

    for i in range(100):
        if f"radius_bb_{i}" in param:
            index = param.index(f"radius_bb_{i}")
            if object_type == "planet":
                param[index] = r"$R_\mathrm{bb,}$" + rf"$_{{{i+1}}}$ ($R_\mathrm{{J}}$)"
            elif object_type == "star":
                param[index] = r"$R_\mathrm{bb,}$" + rf"$_\mathrm{{{i+1}}}$ (au)"

        else:
            break

    if "log_powerlaw_a" in param:
        index = param.index("log_powerlaw_a")
        param[index] = r"$a_\mathrm{powerlaw}$"

    if "log_powerlaw_b" in param:
        index = param.index("log_powerlaw_b")
        param[index] = r"$b_\mathrm{powerlaw}$"

    if "log_powerlaw_c" in param:
        index = param.index("log_powerlaw_c")
        param[index] = r"$c_\mathrm{powerlaw}$"

    if "pt_smooth" in param:
        index = param.index("pt_smooth")
        param[index] = r"$\sigma_\mathrm{P-T}$"

    if "abund_smooth" in param:
        index = param.index("abund_smooth")
        param[index] = r"$\sigma_\mathrm{abund}$"

    if "log_prob" in param:
        index = param.index("log_prob")
        param[index] = r"$\log\,\mathcal{L}$"

    if "ln_prob" in param:
        index = param.index("ln_prob")
        param[index] = r"$\ln\,\mathcal{L}$"

    if "log_tau_cloud" in param:
        index = param.index("log_tau_cloud")
        param[index] = r"$\log\,\tau_\mathrm{cloud}$"

    if "veil_a" in param:
        index = param.index("veil_a")
        param[index] = r"$a_\mathrm{veil}$"

    if "veil_b" in param:
        index = param.index("veil_b")
        param[index] = r"$b_\mathrm{veil}$"

    if "veil_ref" in param:
        index = param.index("veil_ref")
        param[index] = r"$F_\mathrm{ref, veil}$"

    if "gauss_amplitude" in param:
        index = param.index("gauss_amplitude")
        param[index] = r"$a$ (W m$^{-2}$ µm$^{-1}$)"

    if "gauss_mean" in param:
        index = param.index("gauss_mean")
        param[index] = r"$\lambda$ (nm)"

    if "gauss_sigma" in param:
        index = param.index("gauss_sigma")
        param[index] = r"$\sigma$ (nm)"

    if "gauss_amplitude_2" in param:
        index = param.index("gauss_amplitude_2")
        param[index] = r"$a_2$ (W m$^{-2}$ µm$^{-1}$)"

    if "gauss_mean_2" in param:
        index = param.index("gauss_mean_2")
        param[index] = r"$\lambda_2$ (nm)"

    if "gauss_sigma_2" in param:
        index = param.index("gauss_sigma_2")
        param[index] = r"$\sigma_2$ (nm)"

    if "gauss_fwhm" in param:
        index = param.index("gauss_fwhm")
        param[index] = r"FWHM (km s$^{-1}$)"

    if "line_flux" in param:
        index = param.index("line_flux")
        param[index] = r"$F_\mathrm{line}$ (W m$^{-2}$)"

    if "line_luminosity" in param:
        index = param.index("line_luminosity")
        param[index] = r"$L_\mathrm{line}$ ($L_\mathrm{\odot}$)"

    if "log_line_lum" in param:
        index = param.index("log_line_lum")
        param[index] = r"$\log\,L_\mathrm{line}/L_\mathrm{\odot}$"

    if "log_acc_lum" in param:
        index = param.index("log_acc_lum")
        param[index] = r"$\log\,L_\mathrm{acc}/L_\mathrm{\odot}$"

    if "line_eq_width" in param:
        index = param.index("line_eq_width")
        param[index] = r"EW ($\mathrm{\AA}$)"

    if "line_vrad" in param:
        index = param.index("line_vrad")
        param[index] = r"$v_\mathrm{r}$ (km s$^{-1}$)"

    if "log_kappa_0" in param:
        index = param.index("log_kappa_0")
        param[index] = r"$\log\,\kappa_0$"

    if "log_kappa_abs" in param:
        index = param.index("log_kappa_abs")
        param[index] = r"$\log\,\kappa_\mathrm{abs}$"

    if "log_kappa_sca" in param:
        index = param.index("log_kappa_sca")
        param[index] = r"$\log\,\kappa_\mathrm{sca}$"

    if "opa_index" in param:
        index = param.index("opa_index")
        param[index] = r"$\xi$"

    if "opa_abs_index" in param:
        index = param.index("opa_abs_index")
        param[index] = r"$\xi_\mathrm{abs}$"

    if "opa_sca_index" in param:
        index = param.index("opa_sca_index")
        param[index] = r"$\xi_\mathrm{sca}$"

    if "log_p_base" in param:
        index = param.index("log_p_base")
        param[index] = r"$\log\,P_\mathrm{cloud}$"

    if "albedo" in param:
        index = param.index("albedo")
        param[index] = r"$\omega$"

    if "opa_knee" in param:
        index = param.index("opa_knee")
        param[index] = r"$\lambda_\mathrm{R}$ (µm)"

    if "lambda_ray" in param:
        index = param.index("lambda_ray")
        param[index] = r"$\lambda_\mathrm{R}$ (µm)"

    if "mix_length" in param:
        index = param.index("mix_length")
        param[index] = r"$\ell_\mathrm{m}$ ($H_\mathrm{p}$)"

    if "spec_weight" in param:
        index = param.index("spec_weight")
        param[index] = r"w$_\mathrm{spec}$"

    if "log_beta_r" in param:
        index = param.index("log_beta_r")
        param[index] = r"$\log\,\beta_\mathrm{r}$"

    if "log_gamma_r" in param:
        index = param.index("log_gamma_r")
        param[index] = r"$\log\,\gamma_\mathrm{r}$"

    if "gamma_r" in param:
        index = param.index("gamma_r")
        param[index] = r"$\gamma_\mathrm{r}$"

    if "log_kappa_gray" in param:
        index = param.index("log_kappa_gray")
        param[index] = r"$\log\,\kappa_\mathrm{gray}$"

    if "log_cloud_top" in param:
        index = param.index("log_cloud_top")
        param[index] = r"$\log\,P_\mathrm{top}$"

    if "flux_scaling" in param:
        index = param.index("flux_scaling")
        param[index] = r"$a_\mathrm{flux}$"

    if "log_flux_scaling" in param:
        index = param.index("log_flux_scaling")
        param[index] = r"$\log\,a_\mathrm{flux}$"

    if "flux_offset" in param:
        index = param.index("flux_offset")
        param[index] = r"$b_\mathrm{flux}$ (W m$^{-2}$ µm$^{-1}$)"

    return param


@typechecked
def quantity_unit(
    param: List[str], object_type: str
) -> Tuple[List[str], List[Optional[str]], List[str]]:
    """
    Function for creating lists with quantities, units,
    and labels for the fitted parameter.

    Parameters
    ----------
    param : list
        List with parameter names.
    object_type : str
        Object type (``'planet'`` or ``'star'``).

    Returns
    -------
    list
        List with the quantities.
    list
        List with the units.
    list
        List with the parameter labels for plots.
    """

    quantity = []
    unit = []
    label = []

    for item in param:
        item_split = item.split("_")

        if item == "teff":
            quantity.append("teff")
            unit.append("K")
            label.append(r"$T_\mathrm{eff}$")

        elif item == "logg":
            quantity.append("logg")
            unit.append(None)
            label.append(r"$\log g$")

        elif item == "metallicity":
            quantity.append("metallicity")
            unit.append(None)
            label.append("[Fe/H]")

        elif item == "feh":
            quantity.append("feh")
            unit.append(None)
            label.append("[Fe/H]")

        elif item == "fsed":
            quantity.append("fsed")
            unit.append(None)
            label.append(r"$f_\mathrm{sed}$")

        elif item == "c_o_ratio":
            quantity.append("c_o_ratio")
            unit.append(None)
            label.append("C/O")

        elif item == "log_kzz":
            quantity.append("log_kzz")
            unit.append(None)
            label.append(r"$\log\,K_\mathrm{zz}$")

        elif item == "ad_index":
            quantity.append("ad_index")
            unit.append(None)
            label.append(r"$\gamma_\mathrm{ad}$")

        elif item == "log_co_iso":
            quantity.append("log_co_iso")
            unit.append(None)
            label.append(r"$\log\,^{12}\mathrm{CO}/^{13}\mathrm{CO}$")

        elif item == "radius":
            quantity.append("radius")

            if object_type == "planet":
                unit.append(r"$R_\mathrm{J}$")
                label.append(r"$R_\mathrm{p}$")

            elif object_type == "star":
                unit.append(r"$R_\mathrm{\odot}$")
                label.append(r"$R_\ast$")

        elif len(item_split) > 1 and item_split[-1].isdigit():
            param_index = int(item_split[-1])

            if item_split[0] == "teff":
                quantity.append(f"teff_{param_index}")
                unit.append("K")
                label.append(rf"$T_\mathrm{{eff,{param_index+1}}}$")

            elif item_split[0] == "logg":
                quantity.append(f"logg_{param_index}")
                unit.append(None)
                label.append(rf"$\log g_\mathrm{{{param_index+1}}}$")

            elif item_split[0] == "feh":
                quantity.append(f"feh_{param_index}")
                unit.append(None)
                label.append(rf"[Fe/H]$_\mathrm{{{param_index+1}}}$")

            elif item_split[0] == "fsed":
                quantity.append(f"fsed_{param_index}")
                unit.append(None)
                label.append(rf"$f_\mathrm{{sed,{param_index+1}}}$")

            elif item_split[0] == "radius":
                quantity.append(f"radius_{param_index}")

                if object_type == "planet":
                    unit.append(r"$R_\mathrm{J}$")

                elif object_type == "star":
                    unit.append(r"$R_\mathrm{\odot}$")

                label.append(rf"$R_\mathrm{{{param_index+1}}}$")

            elif item[:7] == "ism_ext":
                quantity.append(f"ism_ext_{param_index}")
                unit.append(None)
                label.append(r"$A_{V,}$" + rf"$_\mathrm{{{param_index+1}}}$")

            elif item[:6] == "ext_av":
                quantity.append(f"ext_av_{param_index}")
                unit.append(None)
                label.append(r"$A_{V,}$" + rf"$_\mathrm{{{param_index+1}}}$")

            elif item_split[0] == "disk" and item_split[1] == "teff":
                quantity.append(f"disk_teff_{param_index}")
                unit.append("K")
                label.append(r"$T_\mathrm{disk,}$" + rf"$_\mathrm{{{param_index+1}}}$")

            elif item_split[0] == "disk" and item_split[1] == "radius":
                quantity.append(f"disk_radius_{param_index}")

                if object_type == "planet":
                    unit.append(r"$R_\mathrm{J}$")

                elif object_type == "star":
                    unit.append("au")

                label.append(r"$R_\mathrm{disk,}$" + rf"$_\mathrm{{{param_index+1}}}$")

        elif item == "distance":
            quantity.append("distance")
            unit.append("pc")
            label.append(r"$d$")

        elif item == "parallax":
            quantity.append("parallax")
            unit.append("mas")
            label.append(r"$\varpi$")

        elif item == "mass":
            quantity.append("mass")

            if object_type == "planet":
                unit.append(r"$M_\mathrm{J}$")
                label.append(r"$M$")

            elif object_type == "star":
                unit.append(r"$M_\mathrm{\odot}$")
                label.append(r"$M_\ast$")

        elif item == "luminosity":
            quantity.append("luminosity")

            if object_type == "planet":
                unit.append(None)
                label.append(r"$\log\,L_\mathrm{p}/L_\mathrm{\odot}$")

            elif object_type == "star":
                unit.append(None)
                label.append(r"$\log\,L_\ast/L_\mathrm{\odot}$")

        elif item == "log_lum":
            quantity.append("log_lum")

            if object_type == "planet":
                unit.append(None)
                label.append(r"$\log\,L/L_\mathrm{\odot}$")

            elif object_type == "star":
                unit.append(None)
                label.append(r"$\log\,L/L_\mathrm{\odot}$")

        elif item == "log_lum_atm":
            quantity.append("log_lum_atm")

            if object_type == "planet":
                unit.append(None)
                label.append(r"$\log\,L_\mathrm{p}/L_\mathrm{\odot}$")

            elif object_type == "star":
                unit.append(None)
                label.append(r"$\log\,L_\ast/L_\mathrm{\odot}$")

        elif item == "log_lum_disk":
            quantity.append("log_lum_disk")
            unit.append(None)
            label.append(r"$\log\,L_\mathrm{disk}/L_\mathrm{\odot}$")

        elif item == "ism_ext":
            quantity.append("ism_ext")
            unit.append(None)
            label.append(r"$A_V$")

        elif item == "ext_av":
            quantity.append("ext_av")
            unit.append(None)
            label.append(r"$A_V$")

        elif item == "lognorm_ext":
            quantity.append("lognorm_ext")
            unit.append(None)
            label.append(r"$A_V$")

        elif item == "powerlaw_ext":
            quantity.append("powerlaw_ext")
            unit.append(None)
            label.append(r"$A_V$")

        elif item.startswith("phot_ext_"):
            quantity.append(item)
            unit.append(None)
            filter_name = item[9:].split("/")[1]
            label.append(rf"$A_\mathrm{{{filter_name}}}$")

        elif item == "pt_smooth":
            quantity.append("pt_smooth")
            unit.append(None)
            label.append(r"$\sigma_\mathrm{P-T}$")

        elif item == "abund_smooth":
            quantity.append("abund_smooth")
            unit.append(None)
            label.append(r"$\sigma_\mathrm{abund}$")

        elif item == "disk_teff":
            quantity.append("disk_teff")
            unit.append("K")
            label.append(r"$T_\mathrm{disk}$")

        elif item == "disk_radius":
            quantity.append("disk_radius")

            if object_type == "planet":
                unit.append(r"$R_\mathrm{J}$")

            elif object_type == "star":
                unit.append("au")

            label.append(r"$R_\mathrm{disk}$")

        elif item == "flux_scaling":
            quantity.append("flux_scaling")
            unit.append(None)
            label.append(r"$a_\mathrm{flux}$")

        elif item == "log_flux_scaling":
            quantity.append("log_flux_scaling")
            unit.append(None)
            label.append(r"$\log\,a_\mathrm{flux}$")

        elif item == "flux_offset":
            quantity.append("flux_offset")
            unit.append(r"W m$^{-2}$ µm$^{-1}$")
            label.append(r"$b_\mathrm{flux}$")

        elif item == "rad_vel":
            quantity.append("rad_vel")
            unit.append(r"km s$^{-1}$")
            label.append(r"$v_\mathrm{r}$")

        elif item == "vsini":
            quantity.append("vsini")
            unit.append(r"km s$^{-1}$")
            label.append(r"$v\,\sin\,i$")

    return quantity, unit, label


@typechecked
def field_bounds_ticks(
    field_range: Tuple[str, str],
    check_subclass: bool,
) -> Tuple[np.ndarray, np.ndarray, List[str]]:
    """
    Function for converting the specified field range into boundaries
    and labels for the discrete colorbar that is plotted with a
    color-magnitude or color-color diagram.

    Parameters
    ----------
    field_range : tuple(str, str)
        Range of the discrete colorbar for the field objects. The tuple
        should contain the lower and upper value ('early M', 'late M',
        'early L', 'late L', 'early T', 'late T', 'early Y). Also
        stellar spectral types can be specified.
    check_subclass : bool
        Set to ``True`` if the discrete colorbar should distinguish
        early and late spectral types with different colors or set
        to ``False`` if subclasses should not be distinguished.

    Returns
    -------
    np.ndarray
        Array with the boundaries for the discrete colorbar.
    np.ndarray
        Array with the midpoints for the discrete colorbar.
    list(str)
        List with the tick labels for the discrete colorbar.
    """

    if check_subclass:
        spectral_ranges = [
            "O0-O4",
            "O5-O9",
            "B0-B4",
            "B5-B9",
            "A0-A4",
            "A5-A9",
            "F0-F4",
            "F5-F9",
            "G0-G4",
            "G5-G9",
            "K0-K4",
            "K5-K9",
            "M0-M4",
            "M5-M9",
            "L0-L4",
            "L5-L9",
            "T0-T4",
            "T5-T9",
            "Y1-Y2",
        ]

        if field_range[0] == "early O":
            index_start = 0
        elif field_range[0] == "late O":
            index_start = 1
        elif field_range[0] == "early B":
            index_start = 2
        elif field_range[0] == "late B":
            index_start = 3
        elif field_range[0] == "early A":
            index_start = 4
        elif field_range[0] == "late A":
            index_start = 5
        elif field_range[0] == "early F":
            index_start = 6
        elif field_range[0] == "late F":
            index_start = 7
        elif field_range[0] == "early G":
            index_start = 8
        elif field_range[0] == "late G":
            index_start = 9
        elif field_range[0] == "early K":
            index_start = 10
        elif field_range[0] == "late K":
            index_start = 11
        elif field_range[0] == "early M":
            index_start = 12
        elif field_range[0] == "late M":
            index_start = 13
        elif field_range[0] == "early L":
            index_start = 14
        elif field_range[0] == "late L":
            index_start = 15
        elif field_range[0] == "early T":
            index_start = 16
        elif field_range[0] == "late T":
            index_start = 17
        elif field_range[0] == "early Y":
            index_start = 18

        if field_range[1] == "early O":
            index_end = 1
        elif field_range[1] == "late O":
            index_end = 2
        elif field_range[1] == "early B":
            index_end = 3
        elif field_range[1] == "late B":
            index_end = 4
        elif field_range[1] == "early A":
            index_end = 5
        elif field_range[1] == "late A":
            index_end = 6
        elif field_range[1] == "early F":
            index_end = 7
        elif field_range[1] == "late F":
            index_end = 8
        elif field_range[1] == "early G":
            index_end = 9
        elif field_range[1] == "late G":
            index_end = 10
        elif field_range[1] == "early K":
            index_end = 11
        elif field_range[1] == "late K":
            index_end = 12
        elif field_range[1] == "early M":
            index_end = 13
        elif field_range[1] == "late M":
            index_end = 14
        elif field_range[1] == "early L":
            index_end = 15
        elif field_range[1] == "late L":
            index_end = 16
        elif field_range[1] == "early T":
            index_end = 17
        elif field_range[1] == "late T":
            index_end = 18
        elif field_range[1] == "early Y":
            index_end = 19

    else:
        spectral_ranges = ["O", "B", "A", "F", "G", "K", "M", "L", "T", "Y"]

        if field_range[0] == "O":
            index_start = 0
        elif field_range[0] == "B":
            index_start = 1
        elif field_range[0] == "A":
            index_start = 2
        elif field_range[0] == "F":
            index_start = 3
        elif field_range[0] == "G":
            index_start = 4
        elif field_range[0] == "K":
            index_start = 5
        elif field_range[0] == "M":
            index_start = 6
        elif field_range[0] == "L":
            index_start = 7
        elif field_range[0] == "T":
            index_start = 8
        elif field_range[0] == "Y":
            index_start = 9

        if field_range[1] == "O":
            index_end = 1
        elif field_range[1] == "B":
            index_end = 2
        elif field_range[1] == "A":
            index_end = 3
        elif field_range[1] == "F":
            index_end = 4
        elif field_range[1] == "G":
            index_end = 5
        elif field_range[1] == "K":
            index_end = 6
        elif field_range[1] == "M":
            index_end = 7
        elif field_range[1] == "L":
            index_end = 8
        elif field_range[1] == "T":
            index_end = 9
        elif field_range[1] == "Y":
            index_end = 10

    index_range = index_end - index_start + 1

    bounds = np.linspace(index_start, index_end, index_range)
    ticks = np.linspace(index_start + 0.5, index_end - 0.5, index_range - 1)

    labels = spectral_ranges[index_start:index_end]

    ticks -= bounds[0]
    bounds -= bounds[0]

    return bounds, ticks, labels


@typechecked
def remove_color_duplicates(
    object_names: List[str], empirical_names: np.ndarray
) -> List[int]:
    """ "
    Function for deselecting young/low-gravity objects that will
    already be plotted individually as directly imaged objects.

    Parameters
    ----------
    object_names : list(str)
        List with names of directly imaged planets and brown dwarfs.
    empirical_names : np.ndarray
        Array with names of young/low-gravity objects.

    Returns
    -------
    list
        List with selected indices of the young/low-gravity objects.
    """

    indices = []

    for i, item in enumerate(empirical_names):
        if item == "beta_Pic_b" and "beta Pic b" in object_names:
            continue

        if item == "HR8799b" and "HR 8799 b" in object_names:
            continue

        if item == "HR8799c" and "HR 8799 c" in object_names:
            continue

        if item == "HR8799d" and "HR 8799 d" in object_names:
            continue

        if item == "HR8799e" and "HR 8799 e" in object_names:
            continue

        if item == "kappa_And_B" and "kappa And b" in object_names:
            continue

        if item == "HD1160B" and "HD 1160 B" in object_names:
            continue

        indices.append(i)

    return indices


@typechecked
def create_model_label(
    model_param: Dict[str, float],
    object_type: str,
    model_name: str,
    inc_model_name: bool,
    leg_param: List[str],
    param_fmt: Dict[str, str],
) -> str:
    """ "
    Function for creating a label that includes the parameters of a
    model spectrum. The label is used in the legend of a SED plot.

    Parameters
    ----------
    model_param : dict
        Dictionary with model parameters.
    model_name : str
        Name of the atmospheric model.
    inc_model_name : bool
        Include the model name in the label.
    object_type : str
        Object type ('planet' or 'star') that determines if
        Jupiter or solar units are used.
    leg_param : list(str)
        List with the parameters to include. Apart from atmospheric
        parameters (e.g. 'teff', 'logg', 'radius') also parameters
        such as 'mass' and 'luminosity' can be included. The default
        atmospheric parameters are included in the legend if the
        argument is an empty list.
    param_fmt : dict(str, str)
        Dictionary with formats that will be used for the model
        parameter. The parameters are included in the ``legend``
        when plotting the model spectra. Default formats are
        used if the argument of ``param_fmt`` is set to ``None``.

    Returns
    -------
    str
        Text label that includes the selected parameter names,
        the model values, and the parameter units.
    """

    # data_file = Path(__file__).parents[1].resolve() / "data/model_data/model_data.json"

    # with open(data_file, "r", encoding="utf-8") as json_file:
    #     model_data = json.load(json_file)
    #     model_list = list(model_data.keys())

    # Do not include these parameters by default in the legend

    not_default = ["distance", "parallax", "mass", "luminosity"]

    # Do not include log_lum because log_lum_atm is already include
    # except when the model spectrum is a blackbody from ReadPlanck

    if model_name != "planck":
        not_default.append("log_lum")

    # Use the model parameters if leg_param is empty

    if len(leg_param) == 0:
        leg_param = list(model_param.keys())

        for param_item in not_default:
            if param_item in leg_param:
                # Do not include the not_default parameters
                leg_param.remove(param_item)

    # Remove parameters from the model_param dictionary
    # if they are not included in the leg_param list

    del_keys = []
    for param_item in model_param.keys():
        if param_item not in leg_param:
            del_keys.append(param_item)

    for param_item in del_keys:
        del model_param[param_item]

    # Get parameters keys, units, and labels

    par_key, par_unit, par_label = quantity_unit(
        param=list(model_param.keys()), object_type=object_type
    )

    # Initiate an empty label that will be returned

    label = ""
    # newline = False

    # Optionally include the model name into the label

    if model_name is not None and inc_model_name:
        label += convert_model_name(model_name)

        if len(par_key) > 0:
            label += ": "

    for param_item in leg_param:
        if param_item in par_key:
            param_idx = par_key.index(param_item)
        else:
            continue

        if param_item[:4] == "teff":
            value = f"{model_param[param_item]:{param_fmt['teff']}}"

        elif param_item[:9] == "disk_teff":
            value = f"{model_param[param_item]:{param_fmt['disk_teff']}}"

        elif param_item[:4] == "logg":
            value = f"{model_param[param_item]:{param_fmt['logg']}}"

        elif param_item[:3] == "feh":
            value = f"{model_param[param_item]:{param_fmt['feh']}}"

        elif param_item[:4] == "fsed":
            value = f"{model_param[param_item]:{param_fmt['fsed']}}"

        elif param_item[:7] == "ism_ext":
            value = f"{model_param[param_item]:{param_fmt['ism_ext']}}"

        elif param_item[:6] == "ext_av":
            value = f"{model_param[param_item]:{param_fmt['ext_av']}}"

        elif param_item[:6] == "radius":
            if object_type == "planet":
                value = f"{model_param[param_item]:{param_fmt['radius']}}"
            elif object_type == "star":
                value = f"{model_param[param_item]*constants.R_JUP/constants.R_SUN:{param_fmt['radius']}}"

        elif param_item[:11] == "disk_radius":
            if object_type == "planet":
                value = f"{model_param[param_item]:{param_fmt['disk_radius']}}"
            elif object_type == "star":
                radius_au = model_param[param_item] * constants.R_JUP / constants.AU
                value = f"{radius_au:{param_fmt['disk_radius']}}"

        elif param_item == "mass":
            if object_type == "planet":
                value = f"{model_param[param_item]:{param_fmt['mass']}}"
            elif object_type == "star":
                value = f"{model_param[param_item]*constants.M_JUP/constants.M_SUN:{param_fmt['mass']}}"

        elif param_item == "luminosity":
            value = f"{np.log10(model_param[param_item]):{param_fmt['luminosity']}}"

        elif param_item == "log_lum":
            value = f"{model_param[param_item]:{param_fmt['log_lum']}}"

        elif param_item == "log_lum_atm":
            value = f"{model_param[param_item]:{param_fmt['log_lum_atm']}}"

        elif param_item == "log_lum_disk":
            value = f"{model_param[param_item]:{param_fmt['log_lum_disk']}}"

        elif param_item in param_fmt:
            value = f"{model_param[param_item]:{param_fmt[param_item]}}"

        elif param_item not in param_fmt:
            warnings.warn(
                f"The '{param_item}' parameter is not "
                "found in the dictionary of 'param_fmt'."
            )

            value = f"{model_param[param_item]:.2f}"

        else:
            continue

        # if len(label) > 80 and newline == False:
        #     label += '\n'
        #     newline = True

        if par_unit[param_idx] is None:
            if len(label) > 0 and label[-2] != ":":
                label += ", "

            label += f"{par_label[param_idx]} = {value}"

        else:
            if len(label) > 0 and label[-2] != ":":
                label += ", "

            label += f"{par_label[param_idx]} = {value} {par_unit[param_idx]}"

    return label


def create_param_format(param_fmt: Optional[Dict[str, str]]) -> Dict[str, str]:
    """ "
    Function for creating a dictionary with parameter formats
    that are used in the legend of a plot.

    Parameters
    ----------
    param_fmt : dict(str, str), None
        Dictionary with formats that will be used for the model
        parameter. The parameters are included in the ``legend``
        when plotting the model spectra. Default formats are
        used if the argument of ``param_fmt`` is set to ``None``.

    Returns
    -------
    dict(str, str)
        Output dictionary with parameter formats.
    """

    if param_fmt is None:
        param_fmt = {}

    # Add missing parameter formats

    param_add = ["teff", "disk_teff", "disk_radius"]

    for param_item in param_add:
        if param_item not in param_fmt:
            param_fmt[param_item] = ".0f"

    param_add = [
        "radius",
        "logg",
        "feh",
        "metallicity",
        "fsed",
        "log_kzz",
        "log_co_iso",
        "distance",
        "parallax",
        "mass",
        "ism_ext",
        "ext_av",
        "lognorm_ext",
        "powerlaw_ext",
        "log_flux_scaling",
        "rad_vel",
        "vsini",
    ]

    for param_item in param_add:
        if param_item not in param_fmt:
            param_fmt[param_item] = ".1f"

    param_add = [
        "co",
        "c_o_ratio",
        "ad_index",
        "luminosity",
        "log_lum",
        "log_lum_atm",
        "log_lum_disk",
    ]

    for param_item in param_add:
        if param_item not in param_fmt:
            param_fmt[param_item] = ".2f"

    param_add = ["flux_scaling", "flux_offset"]

    for param_item in param_add:
        if param_item not in param_fmt:
            param_fmt[param_item] = ".2e"

    return param_fmt



"""
Module for adding the IRTF Spectral Library to the database.
"""

from pathlib import Path
from typing import Optional, List

import h5py
import numpy as np
import pandas as pd
import pooch

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import extract_tarfile, update_sptype
from species.util.query_util import get_parallax, get_simbad


@typechecked
def add_irtf(
    input_path: str, database: h5py._hl.files.File, sptypes: Optional[List[str]] = None
) -> None:
    """
    Function for adding the IRTF Spectral Library to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        Database.
    sptypes : list(str), None
        List with the spectral types ('F', 'G', 'K', 'M', 'L', 'T'). All spectral types are
        included if set to ``None``.

    Returns
    -------
    NoneType
        None
    """

    if sptypes is None:
        sptypes = ["F", "G", "K", "M", "L", "T"]

    url = "https://home.strw.leidenuniv.nl/~stolker/species/parallax.dat"
    input_file = "parallax.dat"
    data_file = Path(input_path) / input_file

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash="e2fe0719a919dc98d24627a12f535862a107e473bc67f09298a40ad474cdd491",
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    parallax_data = pd.pandas.read_csv(
        data_file,
        usecols=[0, 1, 2],
        names=["object", "parallax", "parallax_error"],
        delimiter=",",
        dtype={"object": str, "parallax": float, "parallax_error": float},
    )

    data_folder = Path(input_path) / "irtf"

    if not data_folder.exists():
        data_folder.mkdir()

    data_file = {
        "F": Path(input_path) / "irtf/F_fits_091201.tar",
        "G": Path(input_path) / "irtf/G_fits_091201.tar",
        "K": Path(input_path) / "irtf/K_fits_091201.tar",
        "M": Path(input_path) / "irtf/M_fits_091201.tar",
        "L": Path(input_path) / "irtf/L_fits_091201.tar",
        "T": Path(input_path) / "irtf/T_fits_091201.tar",
    }

    data_folder = {
        "F": Path(input_path) / "irtf/F_fits_091201",
        "G": Path(input_path) / "irtf/G_fits_091201",
        "K": Path(input_path) / "irtf/K_fits_091201",
        "M": Path(input_path) / "irtf/M_fits_091201",
        "L": Path(input_path) / "irtf/L_fits_091201",
        "T": Path(input_path) / "irtf/T_fits_091201",
    }

    irtf_folder = Path(input_path) / "irtf/"

    known_hash = {
        "F": "2bc34cfc1262582a8825c4b16d97c47f93cedb46dfcafa43b9db6dfb6fabe9c8",
        "G": "555c19d41dcb5278b796609e6eaceb5d1e8e484cb3b030fda5437722a1a84238",
        "K": "6c5b234a01681c174c3174366e7ee7b90ff35c542f762b9846cedb1d796c50a0",
        "M": "5488656537062593af43c175f9ee5068d0b7dece19b26175cce97d0593232436",
        "L": "2c2f6507c1dca2b81b5f085b3422664d2c21c6e77997037ab053dd77cc2d3381",
        "T": "74675599470c2e86803e9f395ec90fd62f8682a29bcbb281dd1ccd12750033e1",
    }

    url_root = "http://irtfweb.ifa.hawaii.edu/~spex/IRTF_Spectral_Library/Data/"

    url = {
        "F": url_root + "F_fits_091201.tar",
        "G": url_root + "G_fits_091201.tar",
        "K": url_root + "K_fits_091201.tar",
        "M": url_root + "M_fits_091201.tar",
        "L": url_root + "L_fits_091201.tar",
        "T": url_root + "T_fits_091201.tar",
    }

    for spt_item in sptypes:
        if not data_file[spt_item].exists():
            print()

            pooch.retrieve(
                url=url[spt_item],
                known_hash=known_hash[spt_item],
                fname=data_file[spt_item].name,
                path=irtf_folder,
                progressbar=True,
            )

    print("\nUnpacking IRTF Spectral Library...", end="", flush=True)

    for spt_item in sptypes:
        extract_tarfile(str(data_file[spt_item]), str(irtf_folder))

    print(" [DONE]")

    print_message = ""
    print()

    for spt_item in sptypes:
        spec_files = sorted(data_folder[spt_item].glob("*"))

        for file_item in spec_files:
            if file_item.stem[-4:] != "_ext":
                spdata, header = fits.getdata(file_item, header=True)
                spdata = np.transpose(spdata)

                name = header["OBJECT"]
                sptype = header["SPTYPE"]

                if name[-2:] == "AB":
                    name = name[:-2]
                elif name[-3:] == "ABC":
                    name = name[:-3]

                spt_split = sptype.split()

                if spt_item in ["L", "T"] or spt_split[1][0] == "V":
                    empty_message = len(print_message) * " "
                    print(f"\r{empty_message}", end="")

                    print_message = f"Adding spectra... {name}"
                    print(f"\r{print_message}", end="")

                    simbad_id = get_simbad(name)

                    if simbad_id is not None:
                        # For backward compatibility
                        if not isinstance(simbad_id, str):
                            simbad_id = simbad_id.decode("utf-8")

                        par_select = parallax_data[parallax_data["object"] == simbad_id]

                        if not par_select.empty:
                            parallax = (
                                par_select["parallax"],
                                par_select["parallax_error"],
                            )
                        else:
                            simbad_id, parallax = get_parallax(name)

                    else:
                        parallax = (np.nan, np.nan)

                    sptype = update_sptype(np.array([sptype]))[0]

                    dset = database.create_dataset(f"spectra/irtf/{name}", data=spdata)

                    dset.attrs["name"] = str(name).encode()
                    dset.attrs["sptype"] = str(sptype).encode()
                    dset.attrs["simbad"] = str(simbad_id).encode()
                    dset.attrs["parallax"] = parallax[0]
                    dset.attrs["parallax_error"] = parallax[1]

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")



"""
Module for adding young, M- and L-type dwarf spectra from
`Bonnefoy et al. (2014) <https://ui.adsabs.harvard.edu/abs/
2014A%26A...562A.127B/abstract>`_ to the database.
"""

import gzip

from pathlib import Path

import h5py
import numpy as np
import pooch

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import extract_tarfile, remove_directory


@typechecked
def add_bonnefoy2014(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the SINFONI spectra of young, M- and L-type
    dwarfs from `Bonnefoy et al. (2014) <https://ui.adsabs.harvard.
    edu/abs/2014A%26A...562A.127B/abstract>`_ to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        The HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    print_text = "spectra of young M/L type objects from Bonnefoy et al. 2014"

    url = "http://cdsarc.u-strasbg.fr/viz-bin/nph-Cat/tar.gz?J/A+A/562/A127/"
    input_file = "J_A+A_562_A127.tar.gz"
    data_file = Path(input_path) / input_file
    data_folder = Path(input_path) / "bonnefoy+2014/"

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    if data_folder.exists():
        remove_directory(data_folder)

    print(f"\nUnpacking {print_text} (2.3 MB)...", end="", flush=True)
    extract_tarfile(str(data_file), str(data_folder))
    print(" [DONE]")

    spec_dict = {}

    data_file = Path(data_folder) / "stars.dat.gz"

    with gzip.open(data_file, "r") as gzip_file:
        for line in gzip_file:
            name = line[:13].decode().strip()
            files = line[80:].decode().strip().split()
            sptype = line[49:56].decode().strip()

            if name == "NAME 2M1207A":
                name = "2M1207A"

            if len(sptype) == 0:
                sptype = None
            elif "." in sptype:
                sptype = sptype[:4]
            else:
                sptype = sptype[:2]

            if name == "Cha1109":
                sptype = "M9"
            elif name == "DH Tau B":
                sptype = "M9"
            elif name == "TWA 22A":
                sptype = "M6"
            elif name == "TWA 22B":
                sptype = "M6"
            elif name == "CT Cha b":
                sptype = "M9"

            spec_dict[name] = {"name": name, "sptype": sptype, "files": files}

    fits_folder = Path(data_folder) / "sp"

    print_message = ""
    print()

    spec_files = sorted(fits_folder.glob("*"))

    for file_item in spec_files:
        fname_split = file_item.stem.split("_")

        data = fits.getdata(file_item)

        for spec_key, spec_value in spec_dict.items():
            if file_item.name in spec_value["files"]:
                if spec_key == "TWA 22AB":
                    # Binary spectrum
                    continue

                if "JHK" in fname_split:
                    spec_value["JHK"] = data

                elif "J" in fname_split:
                    spec_value["J"] = data

                elif "H+K" in fname_split or "HK" in fname_split:
                    spec_value["HK"] = data

    for spec_key, spec_value in spec_dict.items():
        empty_message = len(print_message) * " "
        print(f"\r{empty_message}", end="")

        print_message = f"Adding spectra... {spec_key}"
        print(f"\r{print_message}", end="")

        if "JHK" in spec_value:
            sp_data = spec_value["JHK"]

        elif "J" in spec_value and "HK" in spec_value:
            sp_data = np.vstack((spec_value["J"], spec_value["HK"]))

        else:
            # Binary spectrum
            continue

        dset = database.create_dataset(
            f"spectra/bonnefoy+2014/{spec_key}", data=sp_data
        )

        dset.attrs["name"] = str(spec_key).encode()
        dset.attrs["sptype"] = str(spec_value["sptype"]).encode()

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")



"""
Module for the photometric data and parallaxes from the JWST late T and
early Y dwarf observations recorded in Beiler et al. 2024 (ApJ).
"""

from pathlib import Path

import h5py
import numpy as np
import pooch

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import update_sptype


@typechecked
def add_jwst_ydwarfs(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the synthesized photometry from 'Precise
    Bolometric Luminosities and Effective Temperatures of 23 Late-T
    and Y Dwarfs Obtained with JWST' by Beiler et al. (2024, ApJ).

    Parameters
    ----------
    input_path : str
        Data folder.
    database : h5py._hl.files.File
        The HDF5 database that has been opened.

    Returns
    -------
    NoneType
        None
    """

    input_file = "Beiler2024_synth_phot.fits"
    data_file = Path(input_path) / "Beiler2024_synth_phot.fits"
    url = "https://home.strw.leidenuniv.nl/~stolker/species/Beiler2024_synth_phot.fits"

    if not data_file.exists():
        print()

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    database.create_group("photometry/beiler2024")

    with fits.open(data_file, mode="update") as hdu_list:
        phot_data = hdu_list[1].data

    parallax = np.array(phot_data["PLX"], dtype=float)  # (mas)
    parallax_error = np.array(phot_data["EPLX"], dtype=float)  # (mas)

    name = np.array(phot_data["Name"]).astype("str")

    sptype_nir = np.array(phot_data["SpT"]).astype("str")
    sptype_nir = np.core.defchararray.strip(sptype_nir)

    sptype = update_sptype(sptype_nir)

    dtype = h5py.special_dtype(vlen=str)

    flag = np.repeat("null", np.size(name))

    dset = database.create_dataset(
        "photometry/beiler2024/flag", (np.size(flag),), dtype=dtype
    )
    dset[...] = flag

    dset = database.create_dataset(
        "photometry/beiler2024/name", (np.size(name),), dtype=dtype
    )
    dset[...] = name

    dset = database.create_dataset(
        "photometry/beiler2024/sptype", (np.size(sptype),), dtype=dtype
    )
    dset[...] = sptype

    database.create_dataset("photometry/beiler2024/parallax", data=parallax)
    database.create_dataset("photometry/beiler2024/parallax_error", data=parallax_error)

    nirc_filts = [
        "F090W",
        "F115W",
        "F140M",
        "F150W",
        "F150W2",
        "F162M",
        "F164N",
        "F182M",
        "F187N",
        "F200W",
        "F210M",
        "F212N",
        "F250M",
        "F277W",
        "F300M",
        "F322W2",
        "F323N",
        "F335M",
        "F356W",
        "F360M",
        "F405N",
        "F410M",
        "F430M",
        "F444W",
        "F460M",
        "F466N",
        "F470N",
        "F480M",
    ]

    for filt in nirc_filts:
        database.create_dataset(
            f"photometry/beiler2024/JWST/NIRCAM.{filt}",
            data=np.array(phot_data[f"{filt}"], dtype=float),
        )

    miri_filts = ["F560W", "F770W", "F1065C", "F1130W", "F1140C"]

    for filt in miri_filts:
        database.create_dataset(
            f"photometry/beiler2024/JWST/MIRI.{filt}",
            data=np.array(phot_data[f"{filt}"], dtype=float),
        )

    database.close()



"""
Module for adding a flux-calibrated spectrum of Vega to the database.
"""

from pathlib import Path

import h5py
import numpy as np
import pooch
import requests

from astropy.io import fits
from typeguard import typechecked


@typechecked
def add_vega(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding a flux-calibrated spectrum of Vega to the
    database. The latest spectrum (alpha_lyr_stis_011.fits) is
    downloaded from the STScI archive (see `CALSPEC page <https://
    www.stsci.edu/hst/instrumentation/reference-data-for-calibration
    -and-tools/astronomical-catalogs/calspec>`_ for details).

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    data_file = Path(input_path) / "alpha_lyr_stis_011.fits"

    if not data_file.exists():
        print()

        try:
            url = (
                "https://archive.stsci.edu/hlsps/reference-atlases"
                "/cdbs/current_calspec/alpha_lyr_stis_011.fits"
            )

            pooch.retrieve(
                url=url,
                known_hash="60aebf5c193223f69061cd176d6309730c3210051fffad0dd6ad44475199ceaa",
                fname="alpha_lyr_stis_011.fits",
                path=input_path,
                progressbar=True,
            )

        except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError):
            url = (
                "https://home.strw.leidenuniv.nl/~stolker/"
                "species/alpha_lyr_stis_011.fits"
            )

            pooch.retrieve(
                url=url,
                known_hash="60aebf5c193223f69061cd176d6309730c3210051fffad0dd6ad44475199ceaa",
                fname="alpha_lyr_stis_011.fits",
                path=input_path,
                progressbar=True,
            )

    if "spectra/calibration/vega" in database:
        del database["spectra/calibration/vega"]

    vega_data = fits.getdata(data_file, ext=1)
    wavelength = vega_data["WAVELENGTH"]  # (Angstrom)
    flux = vega_data["FLUX"]  # (erg s-1 cm-2 A-1)
    error_stat = vega_data["STATERROR"]  # (erg s-1 cm-2 A-1)
    error_sys = vega_data["SYSERROR"]  # (erg s-1 cm-2 A-1)

    wavelength *= 1e-4  # (Angstrom) -> (um)
    flux *= 1e-3 * 1e4  # (erg s-1 cm-2 A-1) -> (W m-2 um-1)
    error_stat *= 1e-3 * 1e4  # (erg s-1 cm-2 A-1) -> (W m-2 um-1)
    error_sys *= 1e-3 * 1e4  # (erg s-1 cm-2 A-1) -> (W m-2 um-1)
    error = np.sqrt(error_stat**2 + error_sys**2)  # (W m-2 um-1)

    print("\nAdding spectrum: Vega", end="", flush=True)

    database.create_dataset(
        "spectra/calibration/vega", data=np.vstack((wavelength, flux, error))
    )

    print("\nReference: Bohlin et al. 2014, PASP, 126")
    print("URL: https://ui.adsabs.harvard.edu/abs/2014PASP..126..711B/abstract")



"""
Module for downloading filter data from the website
of the SVO Filter Profile Service.
"""

import os
import warnings

from urllib.request import urlretrieve
from typing import Optional, Tuple

import h5py
import numpy as np
import pooch

from astropy.io.votable import parse_single_table
from typeguard import typechecked


@typechecked
def download_filter(
    filter_id: str,
    input_path: str,
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray], Optional[str]]:
    """
    Function for downloading filter profile data
    from the SVO Filter Profile Service.

    Parameters
    ----------
    filter_id : str
        Filter name as listed on the website of the `SVO
        Filter Profile Service
        <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    np.ndarray
        Wavelength (um).
    np.ndarray
        Fractional transmission.
    str
        Detector type ('energy' or 'photon').
    """

    if filter_id == "Magellan/VisAO.rp":
        url = "https://xwcl.science/magao/visao/VisAO_rp_filter_curve.dat"
        filter_path = os.path.join(input_path, "VisAO_rp_filter_curve.dat")

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname="VisAO_rp_filter_curve.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission, _, _ = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "Magellan/VisAO.ip":
        url = "https://xwcl.science/magao/visao/VisAO_ip_filter_curve.dat"
        filter_path = os.path.join(input_path, "VisAO_ip_filter_curve.dat")

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname="VisAO_ip_filter_curve.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission, _, _ = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "Magellan/VisAO.zp":
        url = "https://xwcl.science/magao/visao/VisAO_zp_filter_curve.dat"
        filter_path = os.path.join(input_path, "VisAO_zp_filter_curve.dat")

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname="VisAO_zp_filter_curve.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission, _, _ = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "Keck/NIRC2.NB_4.05":
        # The filter profile of Br_alpha has been digitized from
        # https://www2.keck.hawaii.edu/inst/nirc2/filters.html

        url = "https://home.strw.leidenuniv.nl/~stolker/species/Keck_NIRC2.NB_4.05.dat"
        filter_path = os.path.join(input_path, "Keck_NIRC2.NB_4.05.dat")

        pooch.retrieve(
            url=url,
            known_hash="7f7c2de8ee48ab74a03948ef92ecc7f6e1001ada63101ea818619748d36be313",
            fname="Keck_NIRC2.NB_4.05.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "Keck/NIRC.Y":
        # The filter profile of the Y band has been
        # adopted from Hillenbrand et al. (2002)

        url = "https://home.strw.leidenuniv.nl/~stolker/species/Keck_NIRC.Y.dat"
        filter_path = os.path.join(input_path, "Keck_NIRC.Y.dat")

        pooch.retrieve(
            url=url,
            known_hash="0e6a30b42aa927fff599eeb5168260d3fc8af23bb83e4e6eafa78c1881ead4e2",
            fname="Keck_NIRC.Y.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id in ["LCO/VisAO.Ys", "Magellan/VisAO.Ys"]:
        url = "https://xwcl.science/magao/visao/VisAO_Ys_filter_curve.dat"
        filter_path = os.path.join(input_path, "VisAO_Ys_filter_curve.dat")

        pooch.retrieve(
            url=url,
            known_hash=None,
            fname="VisAO_Ys_filter_curve.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission, _, _ = np.loadtxt(filter_path, unpack=True)

        # Remove wavelengths with zero transmission
        wavelength = wavelength[:-7]
        transmission = transmission[:-7]

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ELT/METIS.Lp":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/TC_filter_Lp.dat"
        filter_path = os.path.join(input_path, "TC_filter_Lp.dat")

        pooch.retrieve(
            url=url,
            known_hash="93605adfc6bf322822270afe33f846cabb67aa3375c17678136aa56bcde6130f",
            fname="TC_filter_Lp.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ELT/METIS.Mp":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/TC_filter_Mp.dat"
        filter_path = os.path.join(input_path, "TC_filter_Mp.dat")

        pooch.retrieve(
            url=url,
            known_hash="47688e42388a51c556f163f19670f04dce2e4d7dca426f47651475858ec78f42",
            fname="TC_filter_Mp.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ELT/METIS.N1":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/TC_filter_N1.dat"
        filter_path = os.path.join(input_path, "TC_filter_N1.dat")

        pooch.retrieve(
            url=url,
            known_hash="987fdf58997af2b150bdf7cd9a6578f154b1a21580b5284786d0e8e55c11f4ce",
            fname="TC_filter_N1.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ELT/METIS.Q1":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/TC_filter_Q1.dat"
        filter_path = os.path.join(input_path, "TC_filter_Q1.dat")

        pooch.retrieve(
            url=url,
            known_hash="a9b923638e3a510fe6aa1b4b126619440029ebabe393bd48b1005dd235f7a975",
            fname="TC_filter_Q1.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ALMA/band3":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/alma_band3.dat"
        filter_path = os.path.join(input_path, "alma_band3.dat")

        pooch.retrieve(
            url=url,
            known_hash="a8f7ccc38b09a01827def95bade05f6f0217187d17a98225419a7e38c280df3a",
            fname="alma_band3.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ALMA/band6":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/alma_band6.dat"
        filter_path = os.path.join(input_path, "alma_band6.dat")

        pooch.retrieve(
            url=url,
            known_hash="eac3b06caae1c9895943f63c1b6a78ce1da238fa3cafdffe38235840e36e9015",
            fname="alma_band6.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    elif filter_id == "ALMA/band7":
        url = "https://home.strw.leidenuniv.nl/~stolker/species/alma_band7.dat"
        filter_path = os.path.join(input_path, "alma_band7.dat")

        pooch.retrieve(
            url=url,
            known_hash="f51ec87a417f0033b02d0aaeeb50934299401eae83589112b931adec3f9873d3",
            fname="alma_band7.dat",
            path=input_path,
            progressbar=False,
        )

        wavelength, transmission = np.loadtxt(filter_path, unpack=True)

        det_type = "photon"

        os.remove(filter_path)

    else:
        url = "http://svo2.cab.inta-csic.es/svo/theory/fps/fps.php?ID=" + filter_id
        filter_path = os.path.join(input_path, "filter.xml")
        urlretrieve(url, filter_path)

        try:
            table = parse_single_table(filter_path)

            wavelength = table.array["Wavelength"]
            transmission = table.array["Transmission"]

        except IndexError:
            wavelength = None
            transmission = None
            det_type = None

            warnings.warn(
                f"Filter '{filter_id}' is not available "
                "on the SVO Filter Profile Service."
            )

        except:
            os.remove(filter_path)

            raise ValueError(
                f"The filter data of '{filter_id}' could not "
                "be downloaded. Perhaps the website of the "
                "SVO Filter Profile Service (http://svo2.cab."
                "inta-csic.es/svo/theory/fps/) is not available?"
            )

        if transmission is not None:
            det_type = table.get_field_by_id("DetectorType").value

            # For backward compatibility
            if not isinstance(det_type, str):
                det_type = det_type.decode("utf-8")

            if int(det_type) == 0:
                det_type = "energy"

            elif int(det_type) == 1:
                det_type = "photon"

            else:
                det_type = "photon"

                warnings.warn(
                    f"Detector type, '{det_type}', not "
                    "recognized. Setting detector "
                    "type to photon-counting detector."
                )

            wavelength *= 1e-4  # (um)

        os.remove(filter_path)

    if wavelength is not None:
        indices = []

        for i in range(transmission.size):
            if i == 0 and transmission[i] == 0.0 and transmission[i + 1] == 0.0:
                indices.append(i)

            elif (
                i == transmission.size - 1
                and transmission[i - 1] == 0.0
                and transmission[i] == 0.0
            ):
                indices.append(i)

            elif (
                transmission[i - 1] == 0.0
                and transmission[i] == 0.0
                and transmission[i + 1] == 0.0
            ):
                indices.append(i)

        wavelength = np.delete(wavelength, indices)
        transmission = np.delete(transmission, indices)

        if np.amin(transmission) < 0.0:
            warnings.warn(
                f"The minimum transmission value of {filter_id} is "
                f"smaller than zero ({np.amin(transmission):.2e}). "
                f"Wavelengths with negative transmission "
                f"values will be removed."
            )

            indices = []

            for i, item in enumerate(transmission):
                if item > 0.0:
                    indices.append(i)

            wavelength = wavelength[indices]
            transmission = transmission[indices]

    return wavelength, transmission, det_type


@typechecked
def add_filter_profile(
    input_path: str, database: h5py._hl.files.File, filter_name: str
) -> None:
    """
    Function for downloading and adding a filter profile
    to the HDF5 database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    filter_name : str
        Filter name from the SVO Filter Profile Service (e.g.,
        'Paranal/NACO.Lp') or a user-defined name if a ``filename``
        is specified.

    Returns
    -------
    None
        NoneType
    """

    wavelength, transmission, detector_type = download_filter(filter_name, input_path)

    if wavelength is not None and transmission is not None:
        wavel_new = [wavelength[0]]
        transm_new = [transmission[0]]

        for i in range(wavelength.size - 1):
            if wavelength[i + 1] > wavel_new[-1]:
                # Required for the issue with the Keck/NIRC2.J filter on SVO
                wavel_new.append(wavelength[i + 1])
                transm_new.append(transmission[i + 1])

        dset = database.create_dataset(
            f"filters/{filter_name}", data=np.column_stack((wavel_new, transm_new))
        )

        dset.attrs["det_type"] = str(detector_type)



"""
Module for adding spectral libraries to the database.
"""

from typing import List, Optional

import h5py

from typeguard import typechecked

from species.data.spec_data.spec_allers2013 import add_allers2013
from species.data.spec_data.spec_bonnefoy2014 import add_bonnefoy2014
from species.data.spec_data.spec_irtf import add_irtf
from species.data.spec_data.spec_kesseli2017 import add_kesseli2017
from species.data.spec_data.spec_spex import add_spex
from species.data.spec_data.spec_vega import add_vega


@typechecked
def add_spec_library(
    input_path: str,
    database: h5py._hl.files.File,
    spec_library: str,
    sptypes: Optional[List[str]] = None,
) -> None:
    """
    Function for adding spectral libraries to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    spec_library : str
        Name of the spectral library ('irtf', 'spex', 'kesseli+2017',
        'bonnefoy+2014', 'allers+2013').
    sptypes : list(str), None
        Spectral types ('F', 'G', 'K', 'M', 'L', 'T'). Currently
        only implemented for ``spec_library='irtf'``.

    Returns
    -------
    None
        NoneType
    """

    if spec_library[0:11] == "allers+2013":
        add_allers2013(input_path, database)

    elif spec_library[0:13] == "bonnefoy+2014":
        add_bonnefoy2014(input_path, database)

    elif spec_library[0:5] == "irtf":
        add_irtf(input_path, database, sptypes)

    elif spec_library[0:12] == "kesseli+2017":
        add_kesseli2017(input_path, database)

    elif spec_library[0:5] == "spex":
        add_spex(input_path, database)

    elif spec_library[0:5] == "vega":
        add_vega(input_path, database)

    else:
        raise ValueError(
            f"The spectral library '{spec_library}' is not supported. "
            "Please adjust the argument of 'spec_library'."
        )



"""
Module for setting up species in the working folder.
"""

import json
import socket
import urllib.request

from configparser import ConfigParser
from os import environ
from pathlib import Path, PosixPath, WindowsPath
from typing import Optional, Union

import h5py

from typeguard import typechecked

from .._version import __version__, __version_tuple__


class SpeciesInit:
    """
    Class for initiating species by creating the database and
    configuration file in case they are not present in the working
    folder, and creating the data folder for storage of input data.
    """

    @typechecked
    def __init__(
        self,
        config_file: Optional[Union[str, PosixPath, WindowsPath]] = None,
        database_file: Optional[Union[str, PosixPath, WindowsPath]] = None,
    ) -> None:
        """
        Parameters
        ----------
        config_file : str, PosixPath, WindowsPath, None
            Path to the configuration file that is stored as
            `species_config.ini`. Setting the argument enables
            reading a configuration file at a different path
            than the current working directory. The default
            location (i.e. the working folder of your script or
            notebook) is used by setting the argument to ``None``.
        database_file : str, PosixPath, WindowsPath, None
            Path to the HDF5 database that is stored as
            `species_database.hdf5`. Setting the argument will
            overwrite the database path in the configuration file.
            The path from the configuration file is used if the
            argument of ``database_file`` is set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        print("=======\nspecies\n=======")

        # Check if there is a new version available

        species_version = (
            f"{__version_tuple__[0]}."
            f"{__version_tuple__[1]}."
            f"{__version_tuple__[2]}"
        )

        try:
            pypi_url = "https://pypi.org/pypi/species/json"

            with urllib.request.urlopen(pypi_url, timeout=1.0) as open_url:
                url_content = open_url.read()
                url_data = json.loads(url_content)
                pypi_version = url_data["info"]["version"]

        except (urllib.error.URLError, socket.timeout):
            pypi_version = None

        if pypi_version is not None:
            pypi_split = pypi_version.split(".")
            current_split = species_version.split(".")

            new_major = (pypi_split[0] == current_split[0]) & (
                pypi_split[1] > current_split[1]
            )

            new_minor = (
                (pypi_split[0] == current_split[0])
                & (pypi_split[1] == current_split[1])
                & (pypi_split[2] > current_split[2])
            )

            if new_major | new_minor:
                print(f"\nA new version ({pypi_version}) is available!")
                print("Update species by running:")
                print("pip install --upgrade species")

        print(f"\nVersion: {__version__}")

        working_folder = Path.cwd()
        print(f"Working folder: {working_folder}")

        if config_file is None:
            config_file = working_folder / "species_config.ini"

        elif isinstance(config_file, str):
            config_file = Path(config_file)
            environ["SPECIES_CONFIG"] = str(config_file)

        config = ConfigParser(allow_no_value=True)

        if config_file.exists():
            print(f"\nConfiguration file: {config_file}")

        else:
            print("\nCreating species_config.ini...", end="", flush=True)

            config.add_section("species")
            # config.set('species', '; File with the HDF5 database')
            config.set("species", "database", "species_database.hdf5")
            # config.set('species', '; File with the HDF5 database')
            config.set("species", "data_folder", "./data/")
            # config.set('species', '; File with the HDF5 database')
            config.set("species", "vega_mag", "0.03")

            with open(config_file, "w", encoding="utf-8") as file_obj:
                config.write(file_obj)

            print(" [DONE]")

        config.read(config_file)
        config_update = False

        if database_file is None:
            if "database" in config["species"]:
                database_file = Path(config["species"]["database"])

            else:
                database_file = Path("./species_database.hdf5")
                config.set("species", "database", "species_database.hdf5")
                config_update = True

        else:
            if isinstance(database_file, str):
                config.set("species", "database", database_file)
                database_file = Path(database_file)
            else:
                config.set("species", "database", str(database_file))

            config_update = True

        if "data_folder" in config["species"]:
            data_folder = Path(config["species"]["data_folder"])

        else:
            data_folder = Path("./data/")
            config.set("species", "data_folder", "./data/")
            config_update = True

        if "vega_mag" in config["species"]:
            vega_mag = config["species"]["vega_mag"]

        else:
            vega_mag = 0.03
            config.set("species", "vega_mag", "0.03")
            config_update = True

        if config_update:
            # If condition is needed because the file should
            # not be opened when using MPI and the config
            # file is already present and no need to update
            with open(config_file, "w", encoding="utf-8") as file_obj:
                config.write(file_obj)

        if database_file.exists():
            print(f"Database file: {database_file}")

        else:
            print("Creating species_database.hdf5...", end="", flush=True)
            h5_file = h5py.File(database_file, "w")
            h5_file.close()
            print(" [DONE]")

        try:
            from mpi4py import MPI

            mpi_rank = MPI.COMM_WORLD.Get_rank()
            MPI.COMM_WORLD.Barrier()

        except ImportError:
            mpi_rank = 0

        # Add samples to the database

        if mpi_rank == 0:
            with h5py.File(database_file, "a") as hdf5_file:
                if "configuration" in hdf5_file:
                    del hdf5_file["configuration"]

                config_group = hdf5_file.create_group("configuration")
                config_group.attrs["config_file"] = str(config_file)
                config_group.attrs["database_file"] = str(database_file)
                config_group.attrs["data_folder"] = str(data_folder)

        if data_folder.exists():
            print(f"Data folder: {data_folder}")

        else:
            print("Creating data folder...", end="", flush=True)
            data_folder.mkdir()
            print(" [DONE]")

        print("\nConfiguration settings:")
        print(f"   - Database: {database_file}")
        print(f"   - Data folder: {data_folder}")
        print(f"   - Magnitude of Vega: {vega_mag}")

        try:
            from mpi4py import MPI

            # Rank of this process in a communicator
            mpi_rank = MPI.COMM_WORLD.Get_rank()

            # Number of processes in a communicator
            mpi_size = MPI.COMM_WORLD.Get_size()

            print("\nMultiprocessing: mpi4py installed")
            print(f"Process number {mpi_rank+1:d} out of {mpi_size:d}...")

        except ImportError:
            print("\nMultiprocessing: mpi4py not installed")



from pathlib import Path

import h5py
import pooch

from typeguard import typechecked

from species.data.isochrone_data.iso_manual import add_manual


@typechecked
def add_nextgen(database: h5py._hl.files.File, input_path: str) -> None:
    """
    Function for adding the NextGen isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    url = (
        "https://home.strw.leidenuniv.nl/~stolker/species/"
        "model.NextGen.M-0.0.MKO.Vega"
    )

    iso_tag = "nextgen"

    input_file = url.rsplit("/", maxsplit=1)[-1]
    data_file = Path(input_path) / input_file

    if not data_file.exists():
        pooch.retrieve(
            url=url,
            known_hash="a72aef342a1782553094114dabc3196c0862a0c17ee5c368fbd1b47b5c611363",
            fname=input_file,
            path=input_path,
            progressbar=True,
        )

    add_manual(
        database=database,
        tag=iso_tag,
        file_name=str(data_file),
        model_name="nextgen",
    )



"""
Module with functionalities for retrieving the age and
bulk parameters of one or multiple planets and/or brown
dwarfs in a system.
"""

import configparser
import os
import sys
import warnings

from typing import Dict, List, Optional, Tuple, Union

import h5py
import numpy as np

try:
    import pymultinest

except:
    warnings.warn(
        "PyMultiNest could not be imported. "
        "Perhaps because MultiNest was not built "
        "and/or found at the LD_LIBRARY_PATH "
        "(Linux) or DYLD_LIBRARY_PATH (Mac)?"
    )

from scipy.stats import norm, truncnorm
from tqdm.auto import tqdm
from typeguard import typechecked

from species.core import constants
from species.read.read_isochrone import ReadIsochrone
from species.util.core_util import print_section


class FitEvolution:
    """
    Class for retrieving evolutionary parameters from the bolometric
    luminosity of one or multiple planets in a planetary system.
    Optionally, dynamical mass and/or radius priors can be applied
    A single age is retrieved, so assuming that a difference
    in formation time is negligible at the age of the system.
    """

    @typechecked
    def __init__(
        self,
        evolution_model: str,
        log_lum: Optional[Union[Tuple[float, float], List[Tuple[float, float]]]],
        age_prior: Optional[Tuple[float, float, float]] = None,
        mass_prior: Optional[
            Union[Tuple[float, float], List[Optional[Tuple[float, float]]]]
        ] = None,
        radius_prior: Optional[
            Union[Tuple[float, float], List[Optional[Tuple[float, float]]]]
        ] = None,
        bounds: Optional[
            Dict[
                str,
                Union[
                    Tuple[float, float],
                    Tuple[Optional[Tuple[float, float]], Optional[Tuple[float, float]]],
                    List[Tuple[float, float]],
                ],
            ]
        ] = None,
        interp_method: str = "linear",
    ) -> None:
        """
        Parameters
        ----------
        evolution_model : str
            Database tag of the isochrone data (e.g. 'ames-cond',
            'sonora+0.0', 'atmo-ceq'). When using an incorrect
            argument, and error message is printed that includes
            a list with the isochrone models that are available
            in the current ``species`` database.
        log_lum : tuple(float, float), list(tuple(float, float))
            List with tuples that contain :math:`\\log10{L/L_\\odot}`
            and the related uncertainty for one or multiple objects.
            The list should follow the alphabetical order of companion
            characters (i.e. b, c, d, etc.) to make sure that the
            labels are correctly shown when plotting results.
        age_prior : tuple(float, float, float), None
            Tuple with an optional (asymmetric) normal prior for the
            age (Myr). The tuple should contain three values, for
            example, ``age_prior=(20., -5., +2.)``. The prior is not
            applied if the argument is set to ``None``.
        mass_prior : tuple(float, float), list(tuple(float, float)), None
            Optional list with tuples that contain the (dynamical)
            masses and the related uncertainty for one or multiple
            objects. These masses will be used as normal prior with
            the fit. The order should be identical to ``log_lum``.
            For fitting multiple objects, an item in the list can be
            to ``None`` to not apply the normal prior on a specific
            object.
        radius_prior : tuple(float, float), list(tuple(float, float)), None
            Optional list with tuples that contain the radii (e.g.
            from and SED fit) and the related uncertainty for one
            or multiple objects. These radii will be used as normal
            prior with the fit. The order should be identical to
            ``log_lum``. For fitting multiple objects, an item in
            the list can be to ``None`` to not apply the normal
            prior on a specific object.
        bounds : dict(str, tuple(float, float)), None
            The boundaries that are used for the uniform or
            log-uniform priors. Fixing a parameter is possible by
            providing the same value as lower and upper boundary
            of the parameter (e.g. ``bounds={'y_frac': (0.25, 0.25)``.
        interp_method : str
            Interpolation method for the isochrone data. The argument
            should be either 'linear', for using a linear 2D
            interpolation with `LinearNDInterpolator
            <https://docs.scipy.org/doc/scipy/reference/
            generated/scipy.interpolate.LinearNDInterpolator.html>`_,
            or 'cubic', for using a 2D cubic interpolation with
            `CloughTocher2DInterpolator <https://docs.scipy.org/
            doc/scipy/reference/generated/scipy.interpolate.
            CloughTocher2DInterpolator.html>`_ (default: 'linear').

        Returns
        -------
        NoneType
            None
        """

        print_section("Fit evolutionary model")

        self.evolution_model = evolution_model
        self.log_lum = log_lum
        self.mass_prior = mass_prior
        self.radius_prior = radius_prior
        self.bounds = bounds
        self.age_prior = age_prior
        self.normal_prior = {}
        self.fix_param = {}
        self.interp_method = interp_method

        print(f"Evolution model: {self.evolution_model}")
        print(f"Luminosity log(L/Lsun): {self.log_lum}")

        if isinstance(self.log_lum, tuple):
            self.log_lum = [self.log_lum]

        self.n_planets = len(self.log_lum)

        if self.mass_prior is None:
            self.mass_prior = []
            for i in range(self.n_planets):
                self.mass_prior.append(None)

        elif isinstance(self.mass_prior, tuple):
            self.mass_prior = [self.mass_prior]

        if self.radius_prior is None:
            self.radius_prior = []
            for i in range(self.n_planets):
                self.radius_prior.append(None)

        elif isinstance(self.radius_prior, tuple):
            self.radius_prior = [self.radius_prior]

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = configparser.ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.database_path = config["species"]["database"]

        # Add grid with evolution data

        with h5py.File(self.database_path, "r") as hdf_file:
            found_group = bool("isochrones" in hdf_file)

            if found_group:
                found_model = bool(f"isochrones/{evolution_model}" in hdf_file)
                tag_list = list(hdf_file["isochrones"])
            else:
                found_model = False
                tag_list = None

        if not found_model:
            raise ValueError(
                f"The isochrones of '{evolution_model}' "
                "are not found in the database. Please "
                "add the isochrones by using the "
                "add_isochrones method of Database. The "
                "following isochrone data are found in "
                f"the database: {tag_list}"
            )

        # Model parameters

        self.model_par = ["age"]

        for i in range(self.n_planets):
            self.model_par.append(f"mass_{i}")

        # Read isochrone grid

        self.read_iso = ReadIsochrone(
            tag=self.evolution_model,
            create_regular_grid=False,
            verbose=False,
            interp_method=self.interp_method,
        )

        # Check if the log_lum values are within the
        # available range of the evolutionary grid

        _, _, _, _, grid_log_lum, _, _, _ = self.read_iso._read_data()

        for planet_idx in range(self.n_planets):
            if self.log_lum[planet_idx][0] < np.amin(grid_log_lum):
                warnings.warn(
                    f"The luminosity of the object with index {planet_idx}, "
                    f"log_lum={self.log_lum[planet_idx][0]}, is smaller than "
                    f"the minimum luminosity in the '{self.evolution_model}' "
                    f"grid, log(L/Lsun)={np.amin(grid_log_lum)}."
                )

            if self.log_lum[planet_idx][0] > np.amax(grid_log_lum):
                warnings.warn(
                    f"The luminosity of the object with index {planet_idx}, "
                    f"log_lum={self.log_lum[planet_idx][0]}, is larger than "
                    f"the maximum luminosity in the '{self.evolution_model}' "
                    f"grid, log(L/Lsun)={np.amax(grid_log_lum)}."
                )

        # Prior boundaries

        if self.bounds is not None:
            # Set manual prior boundaries

            bounds_grid = {}
            for param_key, param_value in self.read_iso.grid_points().items():
                if param_key in ["age", "mass"]:
                    bounds_grid[param_key] = (
                        np.amin(param_value),
                        np.amax(param_value),
                    )

            for param_key, param_value in bounds_grid.items():
                for planet_idx in range(self.n_planets):
                    if param_key == "age":
                        param_new = param_key
                    else:
                        param_new = f"{param_key}_{planet_idx}"

                    if param_new not in self.bounds:
                        # Set the parameter boundaries to the grid
                        # boundaries if set to None or not found
                        self.bounds[param_new] = bounds_grid[param_key]

                    else:
                        if self.bounds[param_new][0] < bounds_grid[param_key][0]:
                            warnings.warn(
                                f"The lower bound on {param_new} "
                                f"({self.bounds[param_new][0]}) is smaller than "
                                f"the lower bound from the available "
                                f"evolution model grid "
                                f"({bounds_grid[param_key][0]}). The lower bound "
                                f"of the {param_new} prior will be adjusted to "
                                f"{bounds_grid[param_key][0]}."
                            )

                            self.bounds[param_new] = (
                                bounds_grid[param_key][0],
                                self.bounds[param_new][1],
                            )

                        if self.bounds[param_new][1] > bounds_grid[param_key][1]:
                            warnings.warn(
                                f"The upper bound on {param_new} "
                                f"({self.bounds[param_new][1]}) is larger than the "
                                f"upper bound from the available evolution "
                                f"model grid ({bounds_grid[param_key][1]}). The "
                                f"bound of the {param_new} prior will be adjusted "
                                f"to {bounds_grid[param_key][1]}."
                            )

                            self.bounds[param_new] = (
                                self.bounds[param_new][0],
                                bounds_grid[param_key][1],
                            )

            for i in range(self.n_planets):
                if f"inflate_lbol{i}" in self.bounds:
                    self.model_par.append(f"inflate_lbol{i}")

            for i in range(self.n_planets):
                if f"inflate_mass{i}" in self.bounds:
                    if self.mass_prior[i] is None:
                        warnings.warn(
                            f"The mass_prior with index "
                            f"{i} is set to None so the "
                            f"inflate_mass{i} parameter "
                            f"will be excluded."
                        )

                        del self.bounds[f"inflate_mass{i}"]

                    else:
                        self.model_par.append(f"inflate_mass{i}")

        else:
            # Set all prior boundaries to the grid boundaries

            self.bounds = {}
            for key, value in self.read_iso.grid_points().items():
                if key == "age":
                    self.bounds[key] = (value[0], value[-1])

                elif key == "mass":
                    for i in range(self.n_planets):
                        self.bounds[f"{key}_{i}"] = (value[0], value[-1])

        # Check if parameters are fixed

        del_param = []

        for key, value in self.bounds.items():
            if value[0] == value[1] and value[0] is not None and value[1] is not None:
                self.fix_param[key] = value[0]
                del_param.append(key)

        for item in del_param:
            del self.bounds[item]
            self.model_par.remove(item)

        if self.fix_param:
            print(f"\nFixing {len(self.fix_param)} parameters:")

            for key, value in self.fix_param.items():
                print(f"   - {key} = {value}")

        print(f"\nFitting {len(self.model_par)} parameters:")

        for item in self.model_par:
            print(f"   - {item}")

        # Printing uniform and normal priors

        print("\nUniform priors (min, max):")

        for param_key, param_value in self.bounds.items():
            print(f"   - {param_key} = {param_value}")

        print(f"\nNormal priors (mean, sigma):")
        print(f"   - Age (Myr): {self.age_prior}")
        print(f"   - Mass (Mjup): {self.mass_prior}")
        print(f"   - Radius (Rjup): {self.radius_prior}")

        if len(self.normal_prior) > 0:
            # Not used by the current implementation
            print("\nNormal priors (mean, sigma):")
            for param_key, param_value in self.normal_prior.items():
                if -0.1 < param_value[0] < 0.1:
                    print(
                        f"   - {param_key} = {param_value[0]:.2e} +/- {param_value[1]:.2e}"
                    )
                else:
                    print(
                        f"   - {param_key} = {param_value[0]:.2f} +/- {param_value[1]:.2f}"
                    )

    @typechecked
    def run_multinest(
        self,
        tag: str,
        n_live_points: int = 200,
        output: str = "multinest/",
    ) -> None:
        """
        Function to run the ``PyMultiNest`` wrapper of the
        ``MultiNest`` sampler. While ``PyMultiNest`` can be
        installed with ``pip`` from the PyPI repository,
        ``MultiNest`` has to be built manually. See the
        ``PyMultiNest`` documentation for details:
        http://johannesbuchner.github.io/PyMultiNest/install.html.
        Note that the library path of ``MultiNest`` should be set
        to the environmental variable ``LD_LIBRARY_PATH`` on a
        Linux machine and ``DYLD_LIBRARY_PATH`` on a Mac.
        Alternatively, the variable can be set before
        importing the ``species`` package, for example:

        .. code-block:: python

            >>> import os
            >>> os.environ['DYLD_LIBRARY_PATH'] = '/path/to/MultiNest/lib'
            >>> import species

        Parameters
        ----------
        tag : str
            Database tag where the samples will be stored.
        n_live_points : int
            Number of live points.
        output : str
            Path that is used for the output files from MultiNest.

        Returns
        -------
        NoneType
            None
        """

        print_section("Nested sampling with MultiNest")

        print(f"Database tag: {tag}")
        print(f"Number of live points: {n_live_points}")
        print(f"Output folder: {output}")
        print()

        try:
            from mpi4py import MPI

            mpi_rank = MPI.COMM_WORLD.Get_rank()
            MPI.COMM_WORLD.Barrier()

        except ImportError:
            mpi_rank = 0

        # Create the output folder if required

        if mpi_rank == 0 and not os.path.exists(output):
            os.mkdir(output)

        # Create a dictionary with the cube indices of the parameters

        cube_index = {}
        for i, item in enumerate(self.model_par):
            cube_index[item] = i

        @typechecked
        def ln_prior(cube, n_dim, n_param) -> None:
            """
            Function to transform the unit cube into the parameter
            cube. It is not clear how to pass additional arguments
            to the function, therefore it is placed here.

            Parameters
            ----------
            cube : pymultinest.run.LP_c_double
                Unit cube.
            n_dim : int
                Number of dimensions.
            n_param : int
                Number of parameters.

            Returns
            -------
            NoneType
                None
            """

            for i, key in enumerate(self.model_par):
                if key != "age":
                    obj_idx = int(key.split("_")[-1])

                if key == "age" and self.age_prior is not None:
                    # Asymmetric and truncated normal age prior

                    if cube[cube_index["age"]] < 0.5:
                        # Use lower errorbar on the age
                        # Truncated normal prior

                        # The truncation values are given in number of
                        # standard deviations relative to the mean
                        # of the normal distribution

                        if "age" in self.bounds:
                            a_trunc = (
                                self.bounds["age"][0] - self.age_prior[0]
                            ) / np.abs(self.age_prior[1])

                            b_trunc = (
                                self.bounds["age"][1] - self.age_prior[0]
                            ) / np.abs(self.age_prior[1])

                        else:
                            a_trunc = -self.age_prior[0] / np.abs(self.age_prior[1])
                            b_trunc = np.inf

                        cube[cube_index["age"]] = truncnorm.ppf(
                            cube[cube_index["age"]],
                            a_trunc,
                            b_trunc,
                            loc=self.age_prior[0],
                            scale=np.abs(self.age_prior[1]),
                        )

                    else:
                        # Use upper errorbar on the age

                        if "age" in self.bounds:
                            # Truncated normal prior

                            a_trunc = (
                                self.bounds["age"][0] - self.age_prior[0]
                            ) / self.age_prior[2]

                            b_trunc = (
                                self.bounds["age"][1] - self.age_prior[0]
                            ) / self.age_prior[2]

                            cube[cube_index["age"]] = truncnorm.ppf(
                                cube[cube_index["age"]],
                                a_trunc,
                                b_trunc,
                                loc=self.age_prior[0],
                                scale=self.age_prior[2],
                            )

                        else:
                            # Regular normal prior
                            cube[cube_index["age"]] = norm.ppf(
                                cube[cube_index["age"]],
                                loc=self.age_prior[0],
                                scale=self.age_prior[2],
                            )

                elif key[:4] == "mass" and self.mass_prior[obj_idx] is not None:
                    # Normal mass prior
                    sigma = self.mass_prior[obj_idx][1]

                    if f"inflate_mass{obj_idx}" in self.bounds:
                        sigma += (
                            self.bounds[f"inflate_mass{obj_idx}"][0]
                            + (
                                self.bounds[f"inflate_mass{obj_idx}"][1]
                                - self.bounds[f"inflate_mass{obj_idx}"][0]
                            )
                            * cube[cube_index[f"inflate_mass{obj_idx}"]]
                        )

                    cube[cube_index[f"mass_{obj_idx}"]] = norm.ppf(
                        cube[cube_index[f"mass_{obj_idx}"]],
                        loc=self.mass_prior[obj_idx][0],
                        scale=sigma,
                    )

                else:
                    cube[i] = (
                        self.bounds[key][0]
                        + (self.bounds[key][1] - self.bounds[key][0]) * cube[i]
                    )

        @typechecked
        def ln_like(params, n_dim, n_param) -> Union[float, np.float64]:
            """
            Function for return the log-likelihood for
            the sampled parameter cube.

            Parameters
            ----------
            params : pymultinest.run.LP_c_double
                Cube with physical parameters.
            n_dim : int
                Number of dimensions. This parameter is mandatory
                but not used by the function.
            n_param : int
                Number of parameters. This parameter is mandatory
                but not used by the function.

            Returns
            -------
            float
                Log-likelihood.
            """

            chi_square = 0.0

            for planet_idx in range(self.n_planets):
                param_names = [
                    "age",
                    f"mass_{planet_idx}",
                ]

                if "age" in self.fix_param:
                    age_param = self.fix_param["age"]
                else:
                    age_param = params[cube_index["age"]]

                param_val = []

                for param_item in param_names:
                    if param_item in self.fix_param:
                        param_val.append(self.fix_param[param_item])

                    else:
                        param_val.append(params[cube_index[param_item]])

                if f"inflate_lbol{planet_idx}" in self.bounds:
                    lbol_var = (
                        self.log_lum[planet_idx][1]
                        + params[cube_index[f"inflate_lbol{planet_idx}"]]
                    ) ** 2.0
                else:
                    lbol_var = self.log_lum[planet_idx][1] ** 2

                param_interp = ["log_lum"]
                if self.radius_prior[planet_idx] is not None:
                    param_interp.append("radius")

                iso_box = self.read_iso.get_isochrone(
                    age=age_param,
                    masses=np.array([params[cube_index[f"mass_{planet_idx}"]]]),
                    filters_color=None,
                    filter_mag=None,
                    param_interp=param_interp,
                )

                chi_square += (
                    self.log_lum[planet_idx][0] - iso_box.log_lum[0]
                ) ** 2 / lbol_var

                # Only required when fitting the
                # inflation on the Lbol variance
                chi_square += np.log(2.0 * np.pi * lbol_var)

                if self.mass_prior[planet_idx] is not None:
                    # Only required when fitting the
                    # inflation on the mass variance
                    if f"inflate_mass{planet_idx}" in self.bounds:
                        mass_var = (
                            self.mass_prior[planet_idx][1]
                            + params[cube_index[f"inflate_mass{planet_idx}"]]
                        ) ** 2.0
                    else:
                        mass_var = self.mass_prior[planet_idx][1] ** 2

                    chi_square += np.log(2.0 * np.pi * mass_var)

                # Radius prior
                if self.radius_prior[planet_idx] is not None:
                    chi_square += (
                        self.radius_prior[planet_idx][0] - iso_box.radius[0]
                    ) ** 2 / self.radius_prior[planet_idx][1] ** 2

                # ln_like += -0.5 * weight * (obj_item[0] - phot_flux) ** 2 / phot_var
                # ln_like += -0.5 * weight * np.log(2.0 * np.pi * phot_var)

            if not np.isfinite(chi_square):
                log_like = -np.inf

            else:
                log_like = -0.5 * chi_square

            return log_like

        pymultinest.run(
            ln_like,
            ln_prior,
            len(self.model_par),
            outputfiles_basename=output,
            resume=False,
            n_live_points=n_live_points,
        )

        # Create the Analyzer object
        analyzer = pymultinest.analyse.Analyzer(
            len(self.model_par),
            outputfiles_basename=output,
            verbose=False,
        )

        # Get a dictionary with the ln(Z) and its errors, the
        # individual modes and their parameters quantiles of
        # the parameter posteriors
        sampling_stats = analyzer.get_stats()

        # Nested sampling global log-evidence
        ln_z = sampling_stats["nested sampling global log-evidence"]
        ln_z_error = sampling_stats["nested sampling global log-evidence error"]
        print(f"\nNested sampling global log-evidence: {ln_z:.2f} +/- {ln_z_error:.2f}")

        # Nested importance sampling global log-evidence
        ln_z = sampling_stats["nested importance sampling global log-evidence"]
        ln_z_error = sampling_stats[
            "nested importance sampling global log-evidence error"
        ]
        print(
            f"Nested importance sampling global log-evidence: {ln_z:.2f} +/- {ln_z_error:.2f}"
        )

        # Get the maximum likelihood sample

        best_params = analyzer.get_best_fit()
        max_lnlike = best_params["log_likelihood"]

        print("\nSample with the maximum likelihood:")
        print(f"   - Log-likelihood = {max_lnlike:.2f}")

        for param_idx, param_item in enumerate(best_params["parameters"]):
            if -0.1 < param_item < 0.1:
                print(f"   - {self.model_par[param_idx]} = {param_item:.2e}")
            else:
                print(f"   - {self.model_par[param_idx]} = {param_item:.2f}")

        # Get the posterior samples

        post_samples = analyzer.get_equal_weighted_posterior()

        # Samples and ln(L)

        ln_prob = post_samples[:, -1]
        samples = post_samples[:, :-1]

        # Adding the fixed parameters to the samples

        if self.fix_param:
            samples_tmp = np.copy(samples)
            self.model_par = ["age"]

            for planet_idx in range(self.n_planets):
                self.model_par.append(f"mass_{planet_idx}")

            for planet_idx in range(self.n_planets):
                if f"inflate_lbol{planet_idx}" in self.bounds:
                    self.model_par.append(f"inflate_lbol{planet_idx}")

            for planet_idx in range(self.n_planets):
                if f"inflate_mass{planet_idx}" in self.bounds:
                    self.model_par.append(f"inflate_mass{planet_idx}")

            samples = np.zeros((samples.shape[0], len(self.model_par)))

            for param_idx, param_item in enumerate(self.model_par):
                if param_item in self.fix_param:
                    samples[:, param_idx] = np.full(
                        samples_tmp.shape[0], self.fix_param[param_item]
                    )
                else:
                    samples[:, param_idx] = samples_tmp[:, cube_index[param_item]]

        # Recreate cube_index dictionary because the fix_param
        # parameters have been included in the samples array

        cube_index = {}
        for param_idx, param_item in enumerate(self.model_par):
            cube_index[param_item] = param_idx

        # Add atmospheric parameters: R, Teff, and log(g)

        print("\nExtracting the posteriors of Teff, R, and log(g):")

        radius = np.zeros((samples.shape[0], self.n_planets))
        log_g = np.zeros((samples.shape[0], self.n_planets))
        t_eff = np.zeros((samples.shape[0], self.n_planets))

        for planet_idx in tqdm(range(self.n_planets)):
            for sample_idx in tqdm(range(samples.shape[0]), leave=False):
                age = samples[sample_idx, cube_index["age"]]
                mass = samples[sample_idx, cube_index[f"mass_{planet_idx}"]]

                iso_box = self.read_iso.get_isochrone(
                    age=age,
                    masses=np.array([mass]),
                    filters_color=None,
                    filter_mag=None,
                    param_interp=["log_lum", "teff", "logg", "radius"],
                )

                radius[sample_idx, planet_idx] = iso_box.radius[0]
                log_g[sample_idx, planet_idx] = iso_box.logg[0]
                t_eff[sample_idx, planet_idx] = iso_box.teff[0]

                # l_bol = 10.0 ** iso_box.log_lum[0] * constants.L_SUN

                # t_eff[sample_idx, planet_idx] = (
                #     l_bol
                #     / (
                #         4.0
                #         * np.pi
                #         * (radius[sample_idx, planet_idx] * constants.R_JUP) ** 2
                #         * constants.SIGMA_SB
                #     )
                # ) ** 0.25

        for planet_idx in range(self.n_planets):
            self.model_par.append(f"teff_{planet_idx}")

        for planet_idx in range(self.n_planets):
            self.model_par.append(f"radius_{planet_idx}")

        for planet_idx in range(self.n_planets):
            self.model_par.append(f"logg_{planet_idx}")

        samples = np.hstack((samples, t_eff, radius, log_g))

        # Recreate cube_index dictionary because of
        # derived parameters that were included

        cube_index = {}
        for param_idx, param_item in enumerate(self.model_par):
            cube_index[param_item] = param_idx

        # Remove outliers

        # percent = np.percentile(samples, (1.0, 99.0), axis=0)
        #
        # for i, item in enumerate(percent[0]):
        #     if i == 0:
        #         indices = samples[:, i] < item
        #     else:
        #         indices += samples[:, i] < item
        #
        # samples = samples[~indices, :]
        #
        # for i, item in enumerate(percent[1]):
        #     if i == 0:
        #         indices = samples[:, i] > item
        #     else:
        #         indices += samples[:, i] > item
        #
        # samples = samples[~indices, :]

        # Apply uncertainty inflation

        for planet_idx in range(self.n_planets):
            if f"inflate_lbol{planet_idx}" in self.bounds:
                # sigma_add = np.median(samples[:, cube_index[f"inflate_lbol{planet_idx}"]])
                index_prob = np.argmax(ln_prob)
                sigma_add = samples[index_prob, cube_index[f"inflate_lbol{planet_idx}"]]

                self.log_lum[planet_idx] = (
                    self.log_lum[planet_idx][0],
                    self.log_lum[planet_idx][1] + sigma_add,
                )

        for planet_idx in range(self.n_planets):
            if f"inflate_mass{planet_idx}" in self.bounds:
                # sigma_add = np.median(samples[:, cube_index[f"inflate_lbol{planet_idx}"]])
                index_prob = np.argmax(ln_prob)
                sigma_add = samples[index_prob, cube_index[f"inflate_mass{planet_idx}"]]

                self.mass_prior[planet_idx] = (
                    self.mass_prior[planet_idx][0],
                    self.mass_prior[planet_idx][1] + sigma_add,
                )

        # Set radius_prior to posterior value if argument was None

        # for planet_idx, planet_item in enumerate(self.radius_prior):
        #     if planet_item is None:
        #         radius_samples = samples[:, cube_index[f"radius_{planet_idx}"]]
        #         self.radius_prior[planet_idx] = (
        #             np.mean(radius_samples),
        #             np.std(radius_samples),
        #         )

        # Adjust mass_prior to posterior value

        # for i, item in enumerate(self.mass_prior):
        #     mass_samples = samples[:, cube_index[f"mass_{i}"]]
        #     self.mass_prior[i] = (np.mean(mass_samples), np.std(mass_samples))

        # Adjust radius_prior to posterior value

        # for i, item in enumerate(self.radius_prior):
        #     radius_samples = samples[:, cube_index[f"radius_{i}"]]
        #     self.radius_prior[i] = (np.mean(radius_samples), np.std(radius_samples))

        # Set age_prior to NaN if no prior was provided

        if self.age_prior is None:
            self.age_prior = [np.nan]

        elif "age" in self.fix_param:
            self.age_prior = (self.fix_param["age"], 0.0)

        # Set mass_prior to NaN if no prior was provided

        for planet_idx, planet_item in enumerate(self.mass_prior):
            if f"mass_{planet_idx}" in self.fix_param:
                self.mass_prior[planet_idx] = (
                    self.fix_param[f"mass_{planet_idx}"],
                    0.0,
                )

            elif planet_item is None:
                self.mass_prior[planet_idx] = np.nan

        # Set radius_prior to NaN if no prior was provided

        for planet_idx, planet_item in enumerate(self.radius_prior):
            if f"radius_{planet_idx}" in self.fix_param:
                self.radius_prior[planet_idx] = (
                    self.fix_param[f"radius_{planet_idx}"],
                    0.0,
                )

            elif planet_item is None:
                self.radius_prior[planet_idx] = np.nan

        # Dictionary with attributes that will be stored

        attr_dict = {
            "model_type": "evolution",
            "model_name": self.evolution_model,
            "ln_evidence": (ln_z, ln_z_error),
            "n_planets": self.n_planets,
            "log_lum": self.log_lum,
            "age_prior": self.age_prior,
            "mass_prior": self.mass_prior,
            "radius_prior": self.radius_prior,
        }

        # Get the MPI rank of the process

        try:
            from mpi4py import MPI

            mpi_rank = MPI.COMM_WORLD.Get_rank()
            MPI.COMM_WORLD.Barrier()

        except ImportError:
            mpi_rank = 0

        # Add samples to the database

        if mpi_rank == 0:
            # Writing the samples to the database is only
            # possible when using a single process
            from species.data.database import Database

            species_db = Database()

            species_db.add_samples(
                tag=tag,
                sampler="multinest",
                samples=samples,
                ln_prob=ln_prob,
                modelpar=self.model_par,
                bounds=self.bounds,
                normal_prior=self.normal_prior,
                fixed_param=self.fix_param,
                attr_dict=attr_dict,
            )
